#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Zsolt Foldvari
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2009       Brian Matherly
# Copyright (C) 2010       Peter Landgren
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011-2017  Paul Franklin
# Copyright (C) 2012       Craig Anderson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Report output generator based on Cairo.
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------

from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from math import radians
import re

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug.docgen import (
    BaseDoc,
    TextDoc,
    DrawDoc,
    ParagraphStyle,
    TableCellStyle,
    SOLID,
    FONT_SANS_SERIF,
    FONT_SERIF,
    FONT_MONOSPACE,
    PARA_ALIGN_CENTER,
    PARA_ALIGN_LEFT,
)
from gramps.gen.plug.report import utils
from gramps.gen.errors import PluginError
from gramps.gen.plug.docbackend import CairoBackend
from gramps.gen.utils.image import resize_to_buffer
from gramps.gui.utils import SystemFonts

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

log = logging.getLogger(".libcairodoc")

# -------------------------------------------------------------------------
#
# Pango modules
#
# -------------------------------------------------------------------------
from gi.repository import Pango, PangoCairo

# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------

# each element draws some extra information useful for debugging
DEBUG = False

# ------------------------------------------------------------------------
#
# Font selection
#
# ------------------------------------------------------------------------

_TTF_FREEFONT = {
    FONT_SERIF: "FreeSerif",
    FONT_SANS_SERIF: "FreeSans",
    FONT_MONOSPACE: "FreeMono",
}

_MS_TTFONT = {
    FONT_SERIF: "Times New Roman",
    FONT_SANS_SERIF: "Arial",
    FONT_MONOSPACE: "Courier New",
}

_GNOME_FONT = {
    FONT_SERIF: "Serif",
    FONT_SANS_SERIF: "Sans",
    FONT_MONOSPACE: "Monospace",
}

font_families = _GNOME_FONT


# FIXME debug logging does not work here.
def set_font_families():
    """Set the used font families depending on availability."""
    global font_families

    fonts = SystemFonts()
    family_names = fonts.get_system_fonts()

    fam = [f for f in _TTF_FREEFONT.values() if f in family_names]
    if len(fam) == len(_TTF_FREEFONT):
        font_families = _TTF_FREEFONT
        log.debug("Using FreeFonts: %s" % font_families)
        return

    fam = [f for f in _MS_TTFONT.values() if f in family_names]
    if len(fam) == len(_MS_TTFONT):
        font_families = _MS_TTFONT
        log.debug("Using MS TrueType fonts: %s" % font_families)
        return

    fam = [f for f in _GNOME_FONT.values() if f in family_names]
    if len(fam) == len(_GNOME_FONT):
        font_families = _GNOME_FONT
        log.debug("Using Gnome fonts: %s" % font_families)
        return

    log.debug("No fonts found.")


set_font_families()

# ------------------------------------------------------------------------
#
# Converter functions
#
# ------------------------------------------------------------------------


def fontstyle_to_fontdescription(font_style):
    """Convert a FontStyle instance to a Pango.FontDescription one.

    Font color and underline are not implemented in Pango.FontDescription,
    and have to be set with Pango.Layout.set_attributes(attrlist) method.

    """
    if font_style.get_bold():
        f_weight = Pango.Weight.BOLD
    else:
        f_weight = Pango.Weight.NORMAL

    if font_style.get_italic():
        f_style = Pango.Style.ITALIC
    else:
        f_style = Pango.Style.NORMAL

    font_description = Pango.FontDescription(font_families[font_style.face])
    font_description.set_size(int(round(font_style.get_size() * Pango.SCALE)))
    font_description.set_weight(f_weight)
    font_description.set_style(f_style)

    return font_description


def tabstops_to_tabarray(tab_stops, dpi):
    """Convert a list of tabs given in cm to a Pango.TabArray."""
    tab_array = Pango.TabArray.new(
        initial_size=len(tab_stops), positions_in_pixels=False
    )

    for index, tab_stop in enumerate(tab_stops):
        location = tab_stop * dpi * Pango.SCALE / 2.54
        tab_array.set_tab(index, Pango.TabAlign.LEFT, int(location))

    return tab_array


def raw_length(s):
    """
    Return the length of the raw string after all pango markup has been removed.
    """
    s = re.sub("<.*?>", "", s)
    s = s.replace("&amp;", "&")
    s = s.replace("&lt;", "<")
    s = s.replace("&gt;", ">")
    s = s.replace("&quot;", '"')
    s = s.replace("&apos;", "'")
    return len(s)


###------------------------------------------------------------------------
###
### Table row style
###
###------------------------------------------------------------------------

##class RowStyle(list):
##"""Specifies the format of a table row.

##RowStyle extents the available styles in

##The RowStyle contains the width of each column as a percentage of the
##width of the full row. Note! The width of the row is not known until
##divide() or draw() method is called.

##"""
##def __init__(self):
##self.columns = []

##def set_columns(self, columns):
##"""Set the number of columns.

##@param columns: number of columns that should be used.
##@param type: int

##"""
##self.columns = columns

##def get_columns(self):
##"""Return the number of columns.
##"""
##return self.columns

##def set_column_widths(self, clist):
##"""Set the width of all the columns at once.

##@param clist: list of width of columns in % of the full row.
##@param tyle: list

##"""
##self.columns = len(clist)
##for i in range(self.columns):
##self.colwid[i] = clist[i]

##def set_column_width(self, index, width):
##"""
##Set the width of a specified column to the specified width.

##@param index: column being set (index starts at 0)
##@param width: percentage of the table width assigned to the column
##"""
##self.colwid[index] = width

##def get_column_width(self, index):
##"""
##Return the column width of the specified column as a percentage of
##the entire table width.

##@param index: column to return (index starts at 0)
##"""
##return self.colwid[index]


class FrameStyle:
    """Define the style properties of a Frame.

    - width: Width of the frame in cm.
    - height: Height of the frame in cm.
    - align: Horizontal position to entire page.
             Available values: 'left','center', 'right'.
    - spacing: Tuple of spacing around the frame in cm. Order of values:
               (left, right, top, bottom).

    """

    def __init__(self, width=0, height=0, align="left", spacing=(0, 0, 0, 0)):
        self.width = width
        self.height = height
        self.align = align
        self.spacing = spacing


# ------------------------------------------------------------------------
#
# Document element classes
#
# ------------------------------------------------------------------------


class GtkDocBaseElement:
    """Base of all document elements.

    Support document element structuring and can render itself onto
    a Cairo surface.

    There are two categories of methods:
      1. hierarchy building methods (add_child, get_children, set_parent,
         get_parent);
      2. rendering methods (divide, draw).

    The hierarchy building methods generally don't have to be overridden in
    the subclass, while the rendering methods (divide, draw) must be
    implemented in the subclasses.

    """

    _type = "BASE"
    _allowed_children = []

    def __init__(self, style=None):
        self._parent = None
        self._children = []
        self._style = style

    def get_type(self):
        """Get the type of this element."""
        return self._type

    def set_parent(self, parent):
        """Set the parent element of this element."""
        self._parent = parent

    def get_parent(self):
        """Get the parent element of this element."""
        return self._parent

    def add_child(self, element):
        """Add a child element.

        Returns False if the child cannot be added (e.g. not an allowed type),
        or True otherwise.

        """
        # check if it is an allowed child for this type
        if element.get_type() not in self._allowed_children:
            log.debug(
                "%r is not an allowed child for %r"
                % (element.__class__, self.__class__)
            )
            return False

        # append the child and set its parent
        self._children.append(element)
        element.set_parent(self)
        return True

    def get_children(self):
        """Get the list of children of this element."""
        return self._children

    def get_marks(self):
        """Get the list of index marks for this element."""
        marks = []
        for child in self._children:
            marks.extend(child.get_marks())
        return marks

    def divide(self, layout, width, height, dpi_x, dpi_y):
        """Divide the element into two depending on available space.

        @param layout: pango layout to write on
        @param type: Pango.Layout
        @param width: width of available space for this element
        @param type: device points
        @param height: height of available space for this element
        @param type: device points
        @param dpi_x: the horizontal resolution
        @param type: dots per inch
        @param dpi_y: the vertical resolution
        @param type: dots per inch

        @return: the divided element, and the height of the first part
        @rtype: (GtkDocXXX-1, GtkDocXXX-2), device points

        """
        raise NotImplementedError

    def draw(self, cairo_context, pango_layout, width, dpi_x, dpi_y):
        """Draw itself onto a cairo surface.

        @param cairo_context: context to draw on
        @param type: cairo.Context class
        @param pango_layout: pango layout to write on
        @param type: Pango.Layout class
        @param width: width of available space for this element
        @param type: device points
        @param dpi_x: the horizontal resolution
        @param type: dots per inch
        @param dpi_y: the vertical resolution
        @param type: dots per inch

        @return: height of the element
        @rtype: device points

        """
        raise NotImplementedError


class GtkDocDocument(GtkDocBaseElement):
    """The whole document or a page."""

    _type = "DOCUMENT"
    _allowed_children = [
        "PARAGRAPH",
        "PAGEBREAK",
        "TABLE",
        "IMAGE",
        "FRAME",
        "TOC",
        "INDEX",
    ]

    def draw(self, cairo_context, pango_layout, width, dpi_x, dpi_y):
        x = y = elem_height = 0

        for elem in self._children:
            cairo_context.translate(x, elem_height)
            elem_height = elem.draw(cairo_context, pango_layout, width, dpi_x, dpi_y)
            y += elem_height

        return y

    def has_toc(self):
        for elem in self._children:
            if elem.get_type() == "TOC":
                return True
        return False

    def has_index(self):
        for elem in self._children:
            if elem.get_type() == "INDEX":
                return True
        return False


class GtkDocPagebreak(GtkDocBaseElement):
    """Implement a page break."""

    _type = "PAGEBREAK"
    _allowed_children = []

    def divide(self, layout, width, height, dpi_x, dpi_y):
        return (None, None), 0


class GtkDocTableOfContents(GtkDocBaseElement):
    """Implement a table of contents."""

    _type = "TOC"
    _allowed_children = []

    def divide(self, layout, width, height, dpi_x, dpi_y):
        return (self, None), 0

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        return 0


class GtkDocAlphabeticalIndex(GtkDocBaseElement):
    """Implement an alphabetical index."""

    _type = "INDEX"
    _allowed_children = []

    def divide(self, layout, width, height, dpi_x, dpi_y):
        return (self, None), 0

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        return 0


class GtkDocParagraph(GtkDocBaseElement):
    """Paragraph."""

    _type = "PARAGRAPH"
    _allowed_children = []

    # line spacing is not defined in ParagraphStyle
    spacingfractionfont = 0.2

    def __init__(self, style, leader=None):
        GtkDocBaseElement.__init__(self, style)

        if leader:
            self._text = leader + "\t"
            # FIXME append new tab to the existing tab list
            self._style.set_tabs([-1 * self._style.get_first_indent()])
        else:
            self._text = ""

        self._plaintext = None
        self._attrlist = None

        self._marklist = []

    def add_text(self, text):
        if self._plaintext is not None:
            raise PluginError(
                "CairoDoc: text is already parsed." " You cannot add text anymore"
            )
        self._text = self._text + text

    def add_mark(self, mark):
        """
        Add an index mark to this paragraph
        """
        self._marklist.append((mark, raw_length(self._text)))

    def get_marks(self):
        """
        Return a list of index marks for this paragraph
        """
        return [elem[0] for elem in self._marklist]

    def __set_marklist(self, marklist):
        """
        Internal method to allow for splitting of paragraphs
        """
        self._marklist = marklist

    def __set_plaintext(self, plaintext):
        """
        Internal method to allow for splitting of paragraphs
        """
        if not isinstance(plaintext, str):
            self._plaintext = plaintext.decode("utf-8")
        else:
            self._plaintext = plaintext

    def __set_attrlist(self, attrlist):
        """
        Internal method to allow for splitting of paragraphs
        """
        self._attrlist = attrlist

    def __parse_text(self):
        """
        Parse the markup text. This method will only do this if not
        done already
        """
        if self._plaintext is None:
            parse_ok, self._attrlist, self._plaintext, accel_char = Pango.parse_markup(
                self._text, -1, "\000"
            )

    def divide(self, layout, width, height, dpi_x, dpi_y):
        self.__parse_text()

        l_margin = self._style.get_left_margin() * dpi_x / 2.54
        r_margin = self._style.get_right_margin() * dpi_x / 2.54
        t_margin = self._style.get_top_margin() * dpi_y / 2.54
        b_margin = self._style.get_bottom_margin() * dpi_y / 2.54
        h_padding = self._style.get_padding() * dpi_x / 2.54
        v_padding = self._style.get_padding() * dpi_y / 2.54
        f_indent = self._style.get_first_indent() * dpi_x / 2.54

        # calculate real width available for text
        text_width = width - l_margin - 2 * h_padding - r_margin
        if f_indent < 0:
            text_width -= f_indent
        layout.set_width(int(text_width * Pango.SCALE))

        # set paragraph properties
        layout.set_wrap(Pango.WrapMode.WORD_CHAR)
        layout.set_indent(int(f_indent * Pango.SCALE))
        layout.set_tabs(tabstops_to_tabarray(self._style.get_tabs(), dpi_x))
        #
        align = self._style.get_alignment_text()
        if align == "left":
            layout.set_justify(False)
            layout.set_alignment(Pango.Alignment.LEFT)
        elif align == "right":
            layout.set_justify(False)
            layout.set_alignment(Pango.Alignment.RIGHT)
        elif align == "center":
            layout.set_justify(False)
            layout.set_alignment(Pango.Alignment.CENTER)
        elif align == "justify":
            # We have a problem, in pango, justify works only on full lines,
            # and we need an alignment for the partial lines. We don't know
            # for justify what alignment the user wants however. We assume
            # here LEFT ...
            layout.set_alignment(Pango.Alignment.LEFT)
            layout.set_justify(True)
        else:
            raise ValueError
        #
        font_style = self._style.get_font()
        layout.set_font_description(fontstyle_to_fontdescription(font_style))
        # set line spacing based on font:
        spacing = font_style.get_size() * self.spacingfractionfont
        layout.set_spacing(int(round(spacing * Pango.SCALE)))

        text_height = height - t_margin - 2 * v_padding

        # calculate where to cut the paragraph
        layout.set_text(self._plaintext, -1)
        layout.set_attributes(self._attrlist)
        layout_width, layout_height = layout.get_pixel_size()
        line_count = layout.get_line_count()
        spacing = layout.get_spacing() / Pango.SCALE

        # if all paragraph fits we don't need to cut
        if layout_height - spacing <= text_height:
            paragraph_height = layout_height + spacing + t_margin + (2 * v_padding)
            if height - paragraph_height > b_margin:
                paragraph_height += b_margin
            return (self, None), paragraph_height

        # we need to cut paragraph:

        # 1. if paragraph part of a cell, we do not divide if only small part,
        # of paragraph can be shown, instead move to next page
        if line_count < 4 and self._parent._type == "CELL":
            return (None, self), 0

        lineiter = layout.get_iter()

        linenr = 0
        linerange = lineiter.get_line_yrange()
        # 2. if nothing fits, move to next page without split
        #  there is a spacing above and under the text
        if linerange[1] - linerange[0] + 2.0 * spacing > text_height * Pango.SCALE:
            return (None, self), 0

        # 3. split the paragraph
        startheight = linerange[0]
        endheight = linerange[1]
        splitline = -1
        if lineiter.at_last_line():
            # only one line of text that does not fit
            return (None, self), 0

        while not lineiter.at_last_line():
            # go to next line, see if all fits, if not split
            lineiter.next_line()
            linenr += 1
            linerange = lineiter.get_line_yrange()
            if linerange[1] - startheight + 2.0 * spacing > text_height * Pango.SCALE:
                splitline = linenr
                break
            endheight = linerange[1]
        if splitline == -1:
            print("CairoDoc STRANGE ")
            return (None, self), 0
        # we split at splitline
        # get index of first character which doesn't fit on available height
        layout_line = layout.get_line(splitline)
        index = layout_line.start_index
        # and divide the text, first create the second part
        new_style = ParagraphStyle(self._style)
        new_style.set_top_margin(0)
        # we split a paragraph, text should begin in correct position: no indent
        # as if the paragraph just continues from normal text
        new_style.set_first_indent(0)
        new_paragraph = GtkDocParagraph(new_style)
        # index is in bytecode in the text..
        new_paragraph.__set_plaintext(self._plaintext.encode("utf-8")[index:])
        # now recalculate the attrilist:
        newattrlist = layout.get_attributes().copy()
        newattrlist.filter(self.filterattr, index)

        ##      GTK3 PROBLEM: get_iterator no longer available!!
        ##      REFERENCES:
        ##          http://www.gramps-project.org/bugs/view.php?id=6208
        ##          https://bugzilla.gnome.org/show_bug.cgi?id=646788
        ##          workaround: https://github.com/matasbbb/pitivit/commit/da815339e5ce3631b122a72158ba9ffcc9ee4372
        ##      OLD EASY CODE:
        ##        oldattrlist = newattrlist.get_iterator()
        ##        while oldattrlist.next():
        ##            vals = oldattrlist.get_attrs()
        ##            #print (vals)
        ##            for attr in vals:
        ##                newattr = attr.copy()
        ##                newattr.start_index -= index if newattr.start_index > index \
        ##                                                else 0
        ##                newattr.end_index -= index
        ##                newattrlist.insert(newattr)
        ##      ## START OF WORKAROUND
        oldtext = self._text
        pos = 0
        realpos = 0
        markstarts = []
        # index is in bytecode in the text.. !!
        while pos < index:
            if realpos >= len(oldtext):
                break
            char = oldtext[realpos]
            if char == "<" and oldtext[realpos + 1] != "/":
                # a markup starts
                end = realpos + oldtext[realpos:].find(">") + 1
                markstarts += [oldtext[realpos:end]]
                realpos = end
            elif char == "<":
                # this is the closing tag, we did not stop yet, so remove tag!
                realpos = realpos + oldtext[realpos:].find(">") + 1
                markstarts.pop()
            else:
                pos += len(char.encode("utf-8"))
                realpos += 1
        # now construct the marked up text to use
        newtext = "".join(markstarts)
        newtext += oldtext[realpos:]
        # have it parsed
        parse_ok, newattrlist, _plaintext, accel_char = Pango.parse_markup(
            newtext, -1, "\000"
        )
        ##      ##END OF WORKAROUND
        new_paragraph.__set_attrlist(newattrlist)
        # then update the first one
        self.__set_plaintext(self._plaintext.encode("utf-8")[:index])
        self._style.set_bottom_margin(0)

        # split the list of index marks
        para1 = []
        para2 = []
        for mark, position in self._marklist:
            if position < index:
                para1.append((mark, position))
            else:
                para2.append((mark, position - index))
        self.__set_marklist(para1)
        new_paragraph.__set_marklist(para2)

        paragraph_height = endheight - startheight + spacing + t_margin + 2 * v_padding
        return (self, new_paragraph), paragraph_height

    def filterattr(self, attr, index):
        """callback to filter out attributes in the removed piece at beginning"""
        if attr.start_index > index or (
            attr.start_index < index and attr.end_index > index
        ):
            return False
        return True

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        self.__parse_text()

        l_margin = self._style.get_left_margin() * dpi_x / 2.54
        r_margin = self._style.get_right_margin() * dpi_x / 2.54
        t_margin = self._style.get_top_margin() * dpi_y / 2.54
        b_margin = self._style.get_bottom_margin() * dpi_y / 2.54
        h_padding = self._style.get_padding() * dpi_x / 2.54
        v_padding = self._style.get_padding() * dpi_y / 2.54
        f_indent = self._style.get_first_indent() * dpi_x / 2.54

        # calculate real width available for text
        text_width = width - l_margin - 2 * h_padding - r_margin
        if f_indent < 0:
            text_width -= f_indent
        layout.set_width(int(text_width * Pango.SCALE))

        # set paragraph properties
        layout.set_wrap(Pango.WrapMode.WORD_CHAR)
        layout.set_indent(int(f_indent * Pango.SCALE))
        layout.set_tabs(tabstops_to_tabarray(self._style.get_tabs(), dpi_x))
        #
        align = self._style.get_alignment_text()
        if align == "left":
            layout.set_justify(False)
            layout.set_alignment(Pango.Alignment.LEFT)
        elif align == "right":
            layout.set_justify(False)
            layout.set_alignment(Pango.Alignment.RIGHT)
        elif align == "center":
            layout.set_justify(False)
            layout.set_alignment(Pango.Alignment.CENTER)
        elif align == "justify":
            # We have a problem, in pango, justify works only on full lines,
            # and we need an alignment for the partial lines. We don't know
            # for justify what alignment the user wants however. We assume
            # here LEFT ...
            layout.set_alignment(Pango.Alignment.LEFT)
            layout.set_justify(True)
        #
        font_style = self._style.get_font()
        layout.set_font_description(fontstyle_to_fontdescription(font_style))
        # set line spacing based on font:
        spacing = font_style.get_size() * self.spacingfractionfont
        layout.set_spacing(int(round(spacing * Pango.SCALE)))

        # layout the text
        layout.set_text(self._plaintext, -1)
        layout.set_attributes(self._attrlist)
        layout_width, layout_height = layout.get_pixel_size()

        # render the layout onto the cairo surface
        x = l_margin + h_padding
        if f_indent < 0:
            x += f_indent
        # 3/4 of the spacing is added above the text, 1/4 is added below
        cr.move_to(x, t_margin + v_padding + spacing * 0.75)
        cr.set_source_rgb(*utils.rgb_color(font_style.get_color()))
        PangoCairo.show_layout(cr, layout)

        # calculate the full paragraph height
        height = layout_height + spacing + t_margin + 2 * v_padding + b_margin

        # draw the borders
        if self._style.get_top_border():
            cr.move_to(l_margin, t_margin)
            cr.rel_line_to(width - l_margin - r_margin, 0)
        if self._style.get_right_border():
            cr.move_to(width - r_margin, t_margin)
            cr.rel_line_to(0, height - t_margin - b_margin)
        if self._style.get_bottom_border():
            cr.move_to(l_margin, height - b_margin)
            cr.rel_line_to(width - l_margin - r_margin, 0)
        if self._style.get_left_border():
            cr.move_to(l_margin, t_margin)
            cr.line_to(0, height - t_margin - b_margin)

        cr.set_line_width(1)
        cr.set_source_rgb(0, 0, 0)
        cr.stroke()

        if DEBUG:
            cr.set_line_width(0.1)
            cr.set_source_rgb(1.0, 0, 0)
            cr.rectangle(0, 0, width, height)
            cr.stroke()
            cr.set_source_rgb(0, 0, 1.0)
            cr.rectangle(
                l_margin,
                t_margin,
                width - l_margin - r_margin,
                height - t_margin - b_margin,
            )
            cr.stroke()

        return height


class GtkDocTable(GtkDocBaseElement):
    """Implement a table."""

    _type = "TABLE"
    _allowed_children = ["ROW"]

    def divide(self, layout, width, height, dpi_x, dpi_y):
        # calculate real table width
        table_width = width * self._style.get_width() / 100

        # calculate the height of each row
        table_height = 0
        row_index = 0
        while row_index < len(self._children):
            row = self._children[row_index]
            (r1, r2), row_height = row.divide(layout, table_width, height, dpi_x, dpi_y)
            if r2 is not None:
                # break the table in two parts
                break
            table_height += row_height
            row_index += 1
            height -= row_height

        # divide the table if any row did not fit
        new_table = None
        if row_index < len(self._children):
            new_table = GtkDocTable(self._style)
            # add the split row
            new_table.add_child(r2)
            list(map(new_table.add_child, self._children[row_index + 1 :]))
            del self._children[row_index + 1 :]

        return (self, new_table), table_height

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        # calculate real table width
        table_width = width * self._style.get_width() / 100
        # TODO is a table always left aligned??
        table_height = 0

        # draw all the rows
        for row in self._children:
            cr.save()
            cr.translate(0, table_height)
            row_height = row.draw(cr, layout, table_width, dpi_x, dpi_y)
            cr.restore()
            table_height += row_height

        if DEBUG:
            cr.set_line_width(0.1)
            cr.set_source_rgb(1.0, 0, 0)
            cr.rectangle(0, 0, table_width, table_height)
            cr.stroke()

        return table_height


class GtkDocTableRow(GtkDocBaseElement):
    """Implement a row in a table."""

    _type = "ROW"
    _allowed_children = ["CELL"]

    def divide(self, layout, width, height, dpi_x, dpi_y):
        # the highest cell gives the height of the row
        cell_heights = []
        dividedrow = False
        cell_width_iter = self._style.__iter__()
        new_row = GtkDocTableRow(self._style)
        for cell in self._children:
            cell_width = 0
            for i in range(cell.get_span()):
                cell_width += next(cell_width_iter)
            cell_width = cell_width * width / 100
            (c1, c2), cell_height = cell.divide(
                layout, cell_width, height, dpi_x, dpi_y
            )
            cell_heights.append(cell_height)
            if c2 is None:
                emptycell = GtkDocTableCell(c1._style, c1.get_span())
                new_row.add_child(emptycell)
            else:
                dividedrow = True
                new_row.add_child(c2)

        # save height [inch] of the row to be able to draw exact cell border
        row_height = max(cell_heights)
        self.height = row_height / dpi_y

        # return the new row if dividing was needed
        if dividedrow:
            if row_height == 0:
                for cell in self._children:
                    cell._style.set_top_border(False)
                    cell._style.set_left_border(False)
                    cell._style.set_right_border(False)
            return (self, new_row), row_height
        else:
            return (self, None), row_height

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        cr.save()

        # get the height of this row
        row_height = self.height * dpi_y

        # draw all the cells in the row
        cell_width_iter = self._style.__iter__()
        for cell in self._children:
            cell_width = 0
            for i in range(cell.get_span()):
                cell_width += next(cell_width_iter)
            cell_width = cell_width * width / 100
            cell.draw(cr, layout, cell_width, row_height, dpi_x, dpi_y)
            cr.translate(cell_width, 0)
        cr.restore()

        if DEBUG:
            cr.set_line_width(0.1)
            cr.set_source_rgb(0, 0, 1.0)
            cr.rectangle(0, 0, width, row_height)
            cr.stroke()

        return row_height


class GtkDocTableCell(GtkDocBaseElement):
    """Implement a cell in a table row."""

    _type = "CELL"
    _allowed_children = ["PARAGRAPH", "IMAGE"]

    def __init__(self, style, span=1):
        GtkDocBaseElement.__init__(self, style)
        self._span = span

    def get_span(self):
        return self._span

    def divide(self, layout, width, height, dpi_x, dpi_y):
        h_padding = self._style.get_padding() * dpi_x / 2.54
        v_padding = self._style.get_padding() * dpi_y / 2.54

        # calculate real available width
        width -= 2 * h_padding
        available_height = height

        # calculate height of each child
        cell_height = 0
        new_cell = None
        e2 = None

        childnr = 0
        for child in self._children:
            if new_cell is None:
                (e1, e2), child_height = child.divide(
                    layout, width, available_height, dpi_x, dpi_y
                )
                cell_height += child_height
                available_height -= child_height
                if e2 is not None:
                    # divide the cell
                    new_style = TableCellStyle(self._style)
                    if e1 is not None:
                        new_style.set_top_border(False)
                    new_cell = GtkDocTableCell(new_style, self._span)
                    new_cell.add_child(e2)
                    # then update this cell
                    self._style.set_bottom_border(False)
                if e1 is not None:
                    childnr += 1
            else:
                # cell has been divided
                new_cell.add_child(child)

        self._children = self._children[:childnr]
        # calculate real height
        if cell_height != 0:
            cell_height += 2 * v_padding

        # a cell can't be divided, return the height
        return (self, new_cell), cell_height

    def draw(self, cr, layout, width, cell_height, dpi_x, dpi_y):
        """Draw a cell.

        This draw method is a bit different from the others, as common
        cell height of all cells in a row is also given as parameter.
        This is needed to be able to draw proper vertical borders around
        each cell, i.e. the border should be as long as the highest cell
        in the given row.

        """
        h_padding = self._style.get_padding() * dpi_x / 2.54
        v_padding = self._style.get_padding() * dpi_y / 2.54

        # calculate real available width
        i_width = width - 2 * h_padding

        # draw children
        cr.save()
        cr.translate(h_padding, v_padding)
        for child in self._children:
            child_height = child.draw(cr, layout, i_width, dpi_x, dpi_y)
            cr.translate(0, child_height)
        cr.restore()

        # draw the borders
        if self._style.get_top_border():
            cr.move_to(0, 0)
            cr.rel_line_to(width, 0)
        if self._style.get_right_border():
            cr.move_to(width, 0)
            cr.rel_line_to(0, cell_height)
        if self._style.get_bottom_border():
            cr.move_to(0, cell_height)
            cr.rel_line_to(width, 0)
        if self._style.get_left_border():
            cr.move_to(0, 0)
            cr.line_to(0, cell_height)

        cr.set_line_width(1)
        cr.set_source_rgb(0, 0, 0)
        cr.stroke()

        if DEBUG:
            cr.set_line_width(0.1)
            cr.set_source_rgb(0, 1.0, 0)
            cr.rectangle(0, 0, width, cell_height)
            cr.stroke()

        return cell_height


class GtkDocPicture(GtkDocBaseElement):
    """Implement an image."""

    _type = "IMAGE"
    _allowed_children = []

    def __init__(self, style, filename, width, height, crop=None):
        GtkDocBaseElement.__init__(self, style)
        self._filename = filename
        self._width = width
        self._height = height
        self._crop = crop

    def divide(self, layout, width, height, dpi_x, dpi_y):
        img_width = self._width * dpi_x / 2.54
        img_height = self._height * dpi_y / 2.54

        # image can't be divided, a new page must begin
        # if it can't fit on the current one
        if img_height <= height:
            return (self, None), img_height
        else:
            return (None, self), 0

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        from gi.repository import Gtk, Gdk

        img_width = self._width * dpi_x / 2.54
        img_height = self._height * dpi_y / 2.54

        if self._style == "right":
            l_margin = width - img_width
        elif self._style == "center":
            l_margin = (width - img_width) / 2.0
        else:
            l_margin = 0

        # load the image and get its extents
        pixbuf = resize_to_buffer(self._filename, [img_width, img_height], self._crop)
        pixbuf_width = pixbuf.get_width()
        pixbuf_height = pixbuf.get_height()

        # calculate the scale to fit image into the set extents
        scale = min(img_width / pixbuf_width, img_height / pixbuf_height)

        # draw the image
        cr.save()
        cr.translate(l_margin, 0)
        cr.scale(scale, scale)
        Gdk.cairo_set_source_pixbuf(
            cr,
            pixbuf,
            (img_width / scale - pixbuf_width) / 2,
            (img_height / scale - pixbuf_height) / 2,
        )
        cr.rectangle(0, 0, img_width / scale, img_height / scale)
        ##gcr.set_source_pixbuf(pixbuf,
        ##(img_width - pixbuf_width) / 2,
        ##(img_height - pixbuf_height) / 2)
        ##cr.rectangle(0 , 0, img_width, img_height)
        ##cr.scale(scale, scale)
        cr.fill()
        cr.restore()

        if DEBUG:
            cr.set_line_width(0.1)
            cr.set_source_rgb(1.0, 0, 0)
            cr.rectangle(l_margin, 0, img_width, img_height)
            cr.stroke()

        return img_height


class GtkDocFrame(GtkDocBaseElement):
    """Implement a frame."""

    _type = "FRAME"
    _allowed_children = ["LINE", "POLYGON", "BOX", "TEXT"]

    def divide(self, layout, width, height, dpi_x, dpi_y):
        frame_width = round(self._style.width * dpi_x / 2.54)
        frame_height = round(self._style.height * dpi_y / 2.54)
        t_margin = self._style.spacing[2] * dpi_y / 2.54
        b_margin = self._style.spacing[3] * dpi_y / 2.54

        # frame can't be divided, a new page must begin
        # if it can't fit on the current one
        if frame_height + t_margin + b_margin <= height:
            return (self, None), frame_height + t_margin + b_margin
        elif frame_height + t_margin <= height:
            return (self, None), height
        else:
            return (None, self), 0

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        frame_width = self._style.width * dpi_x / 2.54
        frame_height = self._style.height * dpi_y / 2.54
        l_margin = self._style.spacing[0] * dpi_x / 2.54
        r_margin = self._style.spacing[1] * dpi_x / 2.54
        t_margin = self._style.spacing[2] * dpi_y / 2.54
        b_margin = self._style.spacing[3] * dpi_y / 2.54

        if self._style.align == "left":
            x_offset = l_margin
        elif self._style.align == "right":
            x_offset = width - r_margin - frame_width
        elif self._style.align == "center":
            x_offset = (width - frame_width) / 2.0
        else:
            raise ValueError

        # draw each element in the frame
        cr.save()
        cr.translate(x_offset, t_margin)
        cr.rectangle(0, 0, frame_width, frame_height)
        cr.clip()

        for elem in self._children:
            elem.draw(cr, layout, frame_width, dpi_x, dpi_y)

        cr.restore()

        if DEBUG:
            cr.set_line_width(0.1)
            cr.set_source_rgb(1.0, 0, 0)
            cr.rectangle(x_offset, t_margin, frame_width, frame_height)
            cr.stroke()

        return frame_height + t_margin + b_margin


class GtkDocLine(GtkDocBaseElement):
    """Implement a line."""

    _type = "LINE"
    _allowed_children = []

    def __init__(self, style, x1, y1, x2, y2):
        GtkDocBaseElement.__init__(self, style)
        self._start = (x1, y1)
        self._end = (x2, y2)

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        start = (self._start[0] * dpi_x / 2.54, self._start[1] * dpi_y / 2.54)
        end = (self._end[0] * dpi_x / 2.54, self._end[1] * dpi_y / 2.54)
        line_color = utils.rgb_color(self._style.get_color())

        cr.save()
        cr.set_source_rgb(*line_color)
        cr.set_line_width(self._style.get_line_width())
        # TODO line style
        line_style = self._style.get_line_style()
        if line_style != SOLID:
            cr.set_dash(self._style.get_dash_style(line_style), 0)
        cr.move_to(*start)
        cr.line_to(*end)
        cr.stroke()
        cr.restore()

        return 0


class GtkDocPolygon(GtkDocBaseElement):
    """Implement a line."""

    _type = "POLYGON"
    _allowed_children = []

    def __init__(self, style, path):
        GtkDocBaseElement.__init__(self, style)
        self._path = path

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        path = [(x * dpi_x / 2.54, y * dpi_y / 2.54) for (x, y) in self._path]
        path_start = path.pop(0)
        path_stroke_color = utils.rgb_color(self._style.get_color())
        path_fill_color = utils.rgb_color(self._style.get_fill_color())

        cr.save()
        cr.move_to(*path_start)
        for x, y in path:
            cr.line_to(x, y)
        cr.close_path()
        cr.set_source_rgb(*path_fill_color)
        cr.fill_preserve()
        cr.set_source_rgb(*path_stroke_color)
        cr.set_line_width(self._style.get_line_width())
        # TODO line style
        line_style = self._style.get_line_style()
        if line_style != SOLID:
            cr.set_dash(self._style.get_dash_style(line_style), 0)
        cr.stroke()
        cr.restore()

        return 0


class GtkDocBox(GtkDocBaseElement):
    """Implement a box with optional shadow around it."""

    _type = "BOX"
    _allowed_children = []

    def __init__(self, style, x, y, width, height):
        GtkDocBaseElement.__init__(self, style)
        self._x = x
        self._y = y
        self._width = width
        self._height = height

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        box_x = self._x * dpi_x / 2.54
        box_y = self._y * dpi_y / 2.54
        box_width = self._width * dpi_x / 2.54
        box_height = self._height * dpi_y / 2.54

        box_stroke_color = utils.rgb_color((0, 0, 0))
        box_fill_color = utils.rgb_color(self._style.get_fill_color())
        shadow_color = utils.rgb_color((192, 192, 192))

        cr.save()

        cr.set_line_width(self._style.get_line_width())
        # TODO line style
        line_style = self._style.get_line_style()
        if line_style != SOLID:
            cr.set_dash(self._style.get_dash_style(line_style), 0)

        if self._style.get_shadow():
            shadow_x = box_x + self._style.get_shadow_space() * dpi_x / 2.54
            shadow_y = box_y + self._style.get_shadow_space() * dpi_y / 2.54

            cr.set_source_rgb(*shadow_color)
            cr.rectangle(shadow_x, shadow_y, box_width, box_height)
            cr.fill()

        cr.rectangle(box_x, box_y, box_width, box_height)
        cr.set_source_rgb(*box_fill_color)
        cr.fill_preserve()
        cr.set_source_rgb(*box_stroke_color)
        cr.stroke()

        cr.restore()

        return 0


class GtkDocText(GtkDocBaseElement):
    """Implement a text on graphical reports."""

    _type = "TEXT"
    _allowed_children = []

    # line spacing is not defined in ParagraphStyle
    spacingfractionfont = 0.2

    def __init__(self, style, vertical_alignment, text, x, y, angle=0, mark=None):
        GtkDocBaseElement.__init__(self, style)
        self._align_y = vertical_alignment
        self._text = text
        self._x = x
        self._y = y
        self._angle = angle
        self._marklist = []
        if mark:
            self._marklist = [mark]

    def draw(self, cr, layout, width, dpi_x, dpi_y):
        text_x = self._x * dpi_x / 2.54
        text_y = self._y * dpi_y / 2.54

        # turn off text wrapping
        layout.set_width(-1)

        # set paragraph properties
        align = self._style.get_alignment_text()
        if align == "left":
            layout.set_justify(False)
            layout.set_alignment(Pango.Alignment.LEFT)
        elif align == "right":
            layout.set_justify(False)
            layout.set_alignment(Pango.Alignment.RIGHT)
        elif align == "center":
            layout.set_justify(False)
            layout.set_alignment(Pango.Alignment.CENTER)
        elif align == "justify":
            # We have a problem, in pango, justify works only on full lines,
            # and we need an alignment for the partial lines. We don't know
            # for justify what alignment the user wants however. We assume
            # here CENTER ...
            layout.set_alignment(Pango.Alignment.CENTER)
            layout.set_justify(True)
        else:
            raise ValueError
        #
        font_style = self._style.get_font()
        layout.set_font_description(fontstyle_to_fontdescription(font_style))
        # set line spacing based on font:
        spacing = font_style.get_size() * self.spacingfractionfont
        layout.set_spacing(int(round(spacing * Pango.SCALE)))

        # layout the text
        layout.set_markup(self._text)
        layout_width, layout_height = layout.get_pixel_size()

        # calculate horizontal and vertical alignment shift
        if align == "left":
            align_x = 0
        elif align == "right":
            align_x = -layout_width
        elif align == "center" or align == "justify":
            align_x = -layout_width / 2
        else:
            raise ValueError

        if self._align_y == "top":
            align_y = 0
        elif self._align_y == "center":
            align_y = -layout_height / 2
        elif self._align_y == "bottom":
            align_y = -layout_height
        else:
            raise ValueError

        # render the layout onto the cairo surface
        cr.save()
        cr.translate(text_x, text_y)
        cr.rotate(radians(self._angle))
        cr.move_to(align_x, align_y)
        cr.set_source_rgb(*utils.rgb_color(font_style.get_color()))
        PangoCairo.show_layout(cr, layout)
        cr.restore()

        return layout_height

    def get_marks(self):
        """
        Return the index mark for this text
        """
        return self._marklist


# ------------------------------------------------------------------------
#
# CairoDoc class
#
# ------------------------------------------------------------------------
class CairoDoc(BaseDoc, TextDoc, DrawDoc):
    """Act as an abstract document that can render onto a cairo context.

    Maintains an abstract model of the document. The root of this abstract
    document is self._doc. The model is build via the subclassed BaseDoc, and
    the implemented TextDoc, DrawDoc interface methods.

    It can render the model onto cairo context pages, according to the received
    page style.

    """

    # BaseDoc implementation
    EXT = "pdf"

    def open(self, filename):
        fe = filename.split(".")
        if len(fe) == 1:
            filename = filename + "." + self.EXT
        elif fe[-1] != self.EXT:
            # NOTE: the warning will be bogus
            # if the EXT isn't properly overridden by derived class
            log.warning(
                _(
                    """Mismatch between selected extension %(ext)s and actual format.
 Writing to %(filename)s in format %(impliedext)s."""
                )
                % {"ext": fe[-1], "filename": filename, "impliedext": self.EXT}
            )
        self._backend = CairoBackend(filename)
        self._doc = GtkDocDocument()
        self._active_element = self._doc
        self._pages = []
        self._elements_to_paginate = []
        self._links_error = False

    def close(self):
        self.run()

    # TextDoc implementation

    def page_break(self):
        self._active_element.add_child(GtkDocPagebreak())

    def start_bold(self):
        self.__write_text("<b>", markup=True)

    def end_bold(self):
        self.__write_text("</b>", markup=True)

    def start_superscript(self):
        self.__write_text("<small><sup>", markup=True)

    def end_superscript(self):
        self.__write_text("</sup></small>", markup=True)

    def start_paragraph(self, style_name, leader=None):
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_paragraph_style(style_name)

        new_paragraph = GtkDocParagraph(style, leader)
        self._active_element.add_child(new_paragraph)
        self._active_element = new_paragraph

    def end_paragraph(self):
        self._active_element = self._active_element.get_parent()

    def start_table(self, name, style_name):
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_table_style(style_name)

        new_table = GtkDocTable(style)
        self._active_element.add_child(new_table)
        self._active_element = new_table

        # we need to remember the column width list from the table style.
        # this is an ugly hack, but got no better idea.
        self._active_row_style = list(
            map(style.get_column_width, list(range(style.get_columns())))
        )
        if self.get_rtl_doc():
            self._active_row_style.reverse()

    def end_table(self):
        self._active_element = self._active_element.get_parent()

    def start_row(self):
        new_row = GtkDocTableRow(self._active_row_style)
        self._active_element.add_child(new_row)
        self._active_element = new_row

    def end_row(self):
        if self.get_rtl_doc():
            self._active_element._children.reverse()
        self._active_element = self._active_element.get_parent()

    def start_cell(self, style_name, span=1):
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_cell_style(style_name)

        new_cell = GtkDocTableCell(style, span)
        self._active_element.add_child(new_cell)
        self._active_element = new_cell

    def end_cell(self):
        self._active_element = self._active_element.get_parent()

    def write_styled_note(
        self, styledtext, format, style_name, contains_html=False, links=False
    ):
        """
        Convenience function to write a styledtext to the cairo doc.
        styledtext : assumed a StyledText object to write
        format : = 0 : Flowed, = 1 : Preformatted
        style_name : name of the style to use for default presentation
        contains_html: bool, the backend should not check if html is present.
            If contains_html=True, then the textdoc is free to handle that in
            some way. Eg, a textdoc could remove all tags, or could make sure
            a link is clickable. CairoDoc does nothing different for html notes
        links:  bool, true if URLs should be made clickable
        """
        text = str(styledtext)

        s_tags = styledtext.get_tags()
        # FIXME: following split should be regex to match \n\s*\n instead?
        markuptext = self._backend.add_markup_from_styled(text, s_tags, split="\n\n")

        if format == 1:
            # preformatted, retain whitespace. Cairo retains \n automatically,
            # so use \n\n for paragraph detection
            # FIXME: following split should be regex to match \n\s*\n instead?
            for line in markuptext.split("\n\n"):
                self.start_paragraph(style_name)
                self.__write_text(line, markup=True, links=links)
                self.end_paragraph()
        elif format == 0:
            # flowed
            # FIXME: following split should be regex to match \n\s*\n instead?
            for line in markuptext.split("\n\n"):
                self.start_paragraph(style_name)
                # flowed, normal whitespace goes away, but we keep linebreak
                lines = line.split("\n")
                newlines = []
                for singleline in lines:
                    newlines.append(" ".join(singleline.split()))
                self.__write_text("\n".join(newlines), markup=True, links=links)
                self.end_paragraph()

    def __markup(self, text, markup=None):
        if not markup:
            # We need to escape the text here for later Pango.Layout.set_markup
            # calls. This way we save the markup created by the report
            # The markup in the note editor is not in the text so is not
            # considered. It must be added by pango too
            text = self._backend.ESCAPE_FUNC()(text)
        return text

    def __write_text(self, text, mark=None, markup=False, links=False):
        """
        @param text: text to write.
        @param mark:  IndexMark to use for indexing
        @param markup: True if text already contains markup info.
                       Then text will no longer be escaped
        @param links:  True if URLs should be made clickable
        """
        if links == True:
            import cairo

            if cairo.cairo_version() < 11210 and self._links_error == False:
                # Cairo v1.12 is suppose to be the first version
                # that supports clickable links
                print(
                    """
WARNING: This version of cairo (%s) does NOT support clickable links.
The first version that is suppose to is v1.12.  See the roadmap:

    http://www.cairographics.org/roadmap/

The work around is to save to another format that supports clickable
links (like ODF) and write PDF from that format.
                """
                    % cairo.version
                )
                self._links_error = True

        text = self.__markup(text, markup)

        if mark:
            self._active_element.add_mark(mark)

        self._active_element.add_text(text)

    def write_text(self, text, mark=None, links=False):
        """Write a normal piece of text according to the
           present style
        @param text: text to write.
        @param mark:  IndexMark to use for indexing
        @param links:  True if URLs should be made clickable
        """
        self.__write_text(text, mark, links=links)

    def write_markup(self, text, s_tags, mark=None):
        """
        Writes the text in the current paragraph.  Should only be used after a
        start_paragraph and before an end_paragraph.

        @param text: text to write. The text is assumed to be _not_ escaped
        @param s_tags:  assumed to be list of styledtexttags to apply to the
                        text
        @param mark:  IndexMark to use for indexing
        """
        markuptext = self._backend.add_markup_from_styled(text, s_tags)
        self.__write_text(markuptext, mark=mark, markup=True)

    def add_media(self, name, pos, x_cm, y_cm, alt="", style_name=None, crop=None):
        new_image = GtkDocPicture(pos, name, x_cm, y_cm, crop=crop)
        self._active_element.add_child(new_image)

        if len(alt):
            style_sheet = self.get_style_sheet()

            style = style_sheet.get_paragraph_style(style_name)
            style.set_alignment(PARA_ALIGN_CENTER)
            # Center the caption under the image
            if pos == "right":
                style.set_left_margin(self.get_usable_width() - new_image._width)
            else:
                style.set_right_margin(self.get_usable_width() - new_image._width)
            new_paragraph = GtkDocParagraph(style)
            new_paragraph.add_text("\n".join(alt))
            self._active_element.add_child(new_paragraph)

    def insert_toc(self):
        """
        Insert a Table of Contents at this point in the document.
        """
        self._doc.add_child(GtkDocTableOfContents())

    def insert_index(self):
        """
        Insert an Alphabetical Index at this point in the document.
        """
        self._doc.add_child(GtkDocAlphabeticalIndex())

    # DrawDoc implementation

    def start_page(self):
        # if this is not the first page we need to "close" the previous one
        children = self._doc.get_children()
        if children and children[-1].get_type() != "PAGEBREAK":
            self._doc.add_child(GtkDocPagebreak())

        new_frame_style = FrameStyle(
            width=self.get_usable_width(), height=self.get_usable_height()
        )
        new_frame = GtkDocFrame(new_frame_style)

        self._active_element.add_child(new_frame)
        self._active_element = new_frame

    def end_page(self):
        self._active_element = self._active_element.get_parent()

    def draw_line(self, style_name, x1, y1, x2, y2):
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_draw_style(style_name)

        new_line = GtkDocLine(style, x1, y1, x2, y2)
        self._active_element.add_child(new_line)

    def draw_path(self, style_name, path):
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_draw_style(style_name)

        new_polygon = GtkDocPolygon(style, path)
        self._active_element.add_child(new_polygon)

    def draw_box(self, style_name, text, x, y, w, h, mark=None):
        """@param mark:  IndexMark to use for indexing"""
        # we handle the box and...
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_draw_style(style_name)

        new_box = GtkDocBox(style, x, y, w, h)
        self._active_element.add_child(new_box)

        # ...the text separately
        paragraph_style_name = style.get_paragraph_style()
        if paragraph_style_name:
            paragraph_style = style_sheet.get_paragraph_style(paragraph_style_name)
            paragraph_style.set_alignment(PARA_ALIGN_LEFT)

            # horizontal position of the text is not included in the style,
            # we assume that it is the size of the shadow, or 0.2mm
            if style.get_shadow():
                x_offset = style.get_shadow_space()
            else:
                x_offset = 0.2

            new_text = GtkDocText(
                paragraph_style,
                "center",
                self.__markup(text),
                x + x_offset,
                y + h / 2,
                angle=0,
                mark=mark,
            )
            self._active_element.add_child(new_text)

    def draw_text(self, style_name, text, x, y, mark=None):
        """@param mark:  IndexMark to use for indexing"""
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_draw_style(style_name)
        paragraph_style_name = style.get_paragraph_style()
        paragraph_style = style_sheet.get_paragraph_style(paragraph_style_name)
        paragraph_style.set_alignment(PARA_ALIGN_LEFT)

        new_text = GtkDocText(
            paragraph_style, "top", self.__markup(text), x, y, angle=0, mark=mark
        )
        self._active_element.add_child(new_text)

    def center_text(self, style_name, text, x, y, mark=None):
        """@param mark:  IndexMark to use for indexing"""
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_draw_style(style_name)
        paragraph_style_name = style.get_paragraph_style()
        paragraph_style = style_sheet.get_paragraph_style(paragraph_style_name)
        paragraph_style.set_alignment(PARA_ALIGN_CENTER)

        new_text = GtkDocText(
            paragraph_style, "top", self.__markup(text), x, y, angle=0, mark=mark
        )
        self._active_element.add_child(new_text)

    def rotate_text(self, style_name, text, x, y, angle, mark=None):
        """@param mark:  IndexMark to use for indexing"""
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_draw_style(style_name)
        paragraph_style_name = style.get_paragraph_style()
        paragraph_style = style_sheet.get_paragraph_style(paragraph_style_name)
        paragraph_style.set_alignment(PARA_ALIGN_CENTER)

        new_text = GtkDocText(
            paragraph_style, "center", self.__markup("\n".join(text)), x, y, angle, mark
        )
        self._active_element.add_child(new_text)

    # paginating and drawing interface

    def run(self):
        """Create the physical output from the meta document.

        It must be implemented in the subclasses. The idea is that with
        different subclass different output could be generated:
        e.g. Print, PDF, PS, PNG (which are currently supported by Cairo).

        """
        raise NotImplementedError

    def paginate_document(self, layout, page_width, page_height, dpi_x, dpi_y):
        """Paginate the entire document."""
        while not self.paginate(layout, page_width, page_height, dpi_x, dpi_y):
            pass

    def paginate(self, layout, page_width, page_height, dpi_x, dpi_y):
        """Paginate the meta document in chunks.

        Only one document level element is handled at one run.

        """
        # if first time run than initialize the variables
        if not self._elements_to_paginate:
            self._elements_to_paginate = self._doc.get_children()[:]
            self._pages.append(GtkDocDocument())
            self._available_height = page_height

        # try to fit the next element to current page, divide it if needed
        if not self._elements_to_paginate:
            # this is a self._doc where nothing has been added. Empty page.
            return True
        elem = self._elements_to_paginate.pop(0)
        (e1, e2), e1_h = elem.divide(
            layout, page_width, self._available_height, dpi_x, dpi_y
        )

        # if (part of) it fits on current page add it
        if e1 is not None:
            self._pages[len(self._pages) - 1].add_child(e1)

        # if elem was divided remember the second half to be processed
        if e2 is not None:
            self._elements_to_paginate.insert(0, e2)

        # calculate how much space left on current page
        self._available_height -= e1_h

        # start new page if needed
        if (e1 is None) or (e2 is not None):
            self._pages.append(GtkDocDocument())
            self._available_height = page_height

        return len(self._elements_to_paginate) == 0

    def draw_page(self, page_nr, cr, layout, width, height, dpi_x, dpi_y):
        """Draw a page on a Cairo context."""
        if DEBUG:
            cr.set_line_width(0.1)
            cr.set_source_rgb(0, 1.0, 0)
            cr.rectangle(0, 0, width, height)
            cr.stroke()

        self._pages[page_nr].draw(cr, layout, width, dpi_x, dpi_y)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The TreeModel for the Address list in the Address Tab.
"""

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.datehandler import get_date


# -------------------------------------------------------------------------
#
# AddressModel
#
# -------------------------------------------------------------------------
class AddressModel(Gtk.ListStore):
    """
    AddressModel derives from the ListStore, defining te items in the list
    """

    def __init__(self, obj_list, dbase):
        """
        AddressModel derives from the ListStore, defining te items in the list
        """
        Gtk.ListStore.__init__(
            self, str, str, str, str, str, str, str, str, bool, bool, object
        )
        self.db = dbase
        for obj in obj_list:
            self.append(
                row=[
                    get_date(obj),
                    obj.street,
                    obj.locality,
                    obj.city,
                    obj.state,
                    obj.country,
                    obj.postal,
                    obj.phone,
                    obj.has_citations(),
                    obj.get_privacy(),
                    obj,
                ]
            )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2005  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
PrivacyBase Object class for Gramps.
"""


# -------------------------------------------------------------------------
#
# PrivacyBase Object
#
# -------------------------------------------------------------------------
class PrivacyBase:
    """
    Base class for privacy-aware objects.
    """

    def __init__(self, source=None):
        """
        Initialize a PrivacyBase.

        If the source is not None, then object is initialized from values of
        the source object.

        :param source: Object used to initialize the new object
        :type source: PrivacyBase
        """

        if source:
            self.private = source.private
        else:
            self.private = False

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return self.private

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        self.private = data
        return self

    def set_privacy(self, val):
        """
        Set or clears the privacy flag of the data.

        :param val: value to assign to the privacy flag. True indicates that
            the record is private, False indicates that it is public.
        :type val: bool
        """
        self.private = val

    def get_privacy(self):
        """
        Return the privacy level of the data.

        :returns: True indicates that the record is private
        :rtype: bool
        """
        return self.private

    def _merge_privacy(self, other):
        """
        Merge the privacy level of this object with that of other.

        :returns: Privacy of merged objects.
        :rtype: bool
        """
        self.private = self.private or other.private



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009  Douglas S. Blank <doug.blank@gmail.com>
# Copyright (C) 2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
from html import escape

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.datehandler import get_date
from gramps.gen.lib import EventType
from gramps.gen.utils.db import get_birth_or_fallback, get_death_or_fallback
from gramps.gen.plug.menu import NumberOption, BooleanOption, EnumeratedListOption
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored


# ------------------------------------------------------------------------
#
# PedigreeGramplet class
#
# ------------------------------------------------------------------------
class PedigreeGramplet(Gramplet):
    def init(self):
        self.set_text(_("No Family Tree loaded."))
        self.set_tooltip(_("Move mouse over links for options"))
        self.set_use_markup(True)
        self.max_generations = 100
        self.show_dates = 1
        self.box_mode = "UTF"

    def build_options(self):
        self.add_option(
            NumberOption(_("Max generations"), self.max_generations, 1, 100)
        )
        self.add_option(BooleanOption(_("Show dates"), bool(self.show_dates)))
        elist = EnumeratedListOption(_("Line type"), self.box_mode)
        elist.add_item("UTF", "UTF")
        elist.add_item("ASCII", "ASCII")
        self.add_option(elist)

    def save_options(self):
        self.max_generations = int(self.get_option(_("Max generations")).get_value())
        self.show_dates = int(self.get_option(_("Show dates")).get_value())
        self.box_mode = self.get_option(_("Line type")).get_value()

    def on_load(self):
        if len(self.gui.data) == 3:
            self.max_generations = int(self.gui.data[0])
            self.show_dates = int(self.gui.data[1])
            self.box_mode = self.gui.data[2]  # "ASCII" or "UTF"

    def save_update_options(self, widget=None):
        self.max_generations = int(self.get_option(_("Max generations")).get_value())
        self.show_dates = int(self.get_option(_("Show dates")).get_value())
        self.box_mode = self.get_option(_("Line type")).get_value()
        self.gui.data = [self.max_generations, self.show_dates, self.box_mode]
        self.update()

    def db_changed(self):
        """
        If a person or family changes, the ancestors of active person might have
        changed.
        """
        self.connect(self.dbstate.db, "person-add", self.update)
        self.connect(self.dbstate.db, "person-delete", self.update)
        self.connect(self.dbstate.db, "family-add", self.update)
        self.connect(self.dbstate.db, "family-delete", self.update)
        self.connect(self.dbstate.db, "person-rebuild", self.update)
        self.connect(self.dbstate.db, "family-rebuild", self.update)

    def active_changed(self, handle):
        self.update()

    def get_boxes(self, generation, what):
        retval = ""
        if self.box_mode == "UTF":
            space = "  "
        elif self.box_mode == "ASCII":
            space = "    "
        space_len = len(space) + 2
        for i in range(generation + 1):
            if self._boxes[i]:
                retval += space + "|"
            else:
                retval += space + " "
        if retval[-1] == " ":
            if what == "sf":
                retval = retval[:-space_len] + "/"
            elif what == "sm":
                retval = retval[:-space_len] + "\\"
        elif retval.endswith("|" + space + "|"):
            retval = retval[:-space_len] + "+"
        if self.box_mode == "UTF":
            retval += "-"
            retval = retval.replace("\\", "\u2514")
            retval = retval.replace("-", "\u2500")
            retval = retval.replace("|", "\u2502")
            retval = retval.replace("/", "\u250c")
        elif self.box_mode == "ASCII":
            retval += "--"
        return retval

    def set_box(self, pos, value):
        self._boxes[pos] = value

    def process_person(self, handle, generation, what):
        if generation > self.max_generations:
            return
        person = self.dbstate.db.get_person_from_handle(handle)
        family_list = person.get_parent_family_handle_list()
        if what == "f":
            if len(family_list) > 0:
                family = self.dbstate.db.get_family_from_handle(family_list[0])
                father = family.get_father_handle()
                mother = family.get_mother_handle()
                if father:
                    self.process_person(father, generation + 1, "f")
                    self.set_box(generation, 1)
                    self.process_person(father, generation + 1, "sf")
                    self.process_person(father, generation + 1, "m")
                elif mother:
                    self.set_box(generation, 1)
        elif what[0] == "s":
            boxes = self.get_boxes(generation, what)
            if what[-1] == "f":
                if self.box_mode == "UTF":
                    boxes = boxes.replace("+", "\u250c")
                else:
                    boxes = boxes.replace("+", "/")
            else:
                if self.box_mode == "UTF":
                    boxes = boxes.replace("+", "\u2514")
                else:
                    boxes = boxes.replace("+", "\\")
            self.append_text(boxes)
            self.link(
                name_displayer.display_name(person.get_primary_name()),
                "Person",
                person.handle,
                tooltip=_("Click to make active\n") + _("Right-click to edit"),
            )
            if self.show_dates:
                self.append_text(" ")
                self.render_text(self.info_string(person))
            self.append_text("\n")
            if generation not in self._generations:
                self._generations[generation] = []
            self._generations[generation].append(handle)
        elif what == "a":
            if self.box_mode == "UTF":
                self.append_text("o" + ("\u2500" * 3))
            elif self.box_mode == "ASCII":
                self.append_text("o---")
            self.append_text(
                "%s  " % name_displayer.display_name(person.get_primary_name())
            )
            if self.show_dates:
                self.render_text(self.info_string(person))
            self.append_text("\n")
            if generation not in self._generations:
                self._generations[generation] = []
            self._generations[generation].append(handle)
        elif what == "m":
            if len(family_list) > 0:
                family = self.dbstate.db.get_family_from_handle(family_list[0])
                mother = family.get_mother_handle()
                if mother:
                    self.process_person(mother, generation + 1, "f")
                    self.process_person(mother, generation + 1, "sm")
                    self.set_box(generation, 0)
                    self.process_person(mother, generation + 1, "m")
            self.set_box(generation, 0)  # regardless, turn off line if on

    def info_string(self, person):
        birth = get_birth_or_fallback(self.dbstate.db, person)
        if birth and birth.get_type() != EventType.BIRTH:
            sdate = get_date(birth)
            if sdate:
                bdate = "<i>%s</i>" % escape(sdate)
            else:
                bdate = ""
        elif birth:
            bdate = escape(get_date(birth))
        else:
            bdate = ""

        death = get_death_or_fallback(self.dbstate.db, person)
        if death and death.get_type() != EventType.DEATH:
            sdate = get_date(death)
            if sdate:
                ddate = "<i>%s</i>" % escape(sdate)
            else:
                ddate = ""
        elif death:
            ddate = escape(get_date(death))
        else:
            ddate = ""

        if bdate and ddate:
            value = _("(b. %(birthdate)s, d. %(deathdate)s)") % {
                "birthdate": bdate,
                "deathdate": ddate,
            }
        elif bdate:
            value = _("(b. %s)") % (bdate)
        elif ddate:
            value = _("(d. %s)") % (ddate)
        else:
            value = ""
        return value

    def main(self):  # return false finishes
        """
        Generator which will be run in the background.
        """
        self._boxes = [0] * (self.max_generations + 1)
        self._generations = {}
        self.gui.buffer.set_text("")
        active_handle = self.get_active("Person")
        if active_handle is None:
            return
        active_person = self.dbstate.db.get_person_from_handle(active_handle)
        # no wrap in Gramplet
        self.no_wrap()
        if active_person is None:
            return
        self.process_person(active_person.handle, 1, "f")  # father
        self.process_person(active_person.handle, 0, "a")  # active #FIXME: should be 1?
        self.process_person(active_person.handle, 1, "m")  # mother
        gens = sorted(self._generations)
        self.append_text(_("\nBreakdown by generation:\n"))
        all = [active_person.handle]
        percent_sign = _("%", "percent sign or text string")
        for g in gens:
            yield True
            count = len(self._generations[g])
            handles = self._generations[g]
            self.append_text("     ")
            if g == 0:
                self.link(
                    _("Generation 1"),
                    "PersonList",
                    handles,
                    tooltip=_("Double-click to see people in generation"),
                )
                percent = glocale.format_string("%.2f", 100) + percent_sign
                self.append_text(
                    _(" has 1 of 1 individual (%(percent)s complete)\n")
                    % {"percent": percent}
                )
            else:
                all.extend(handles)
                self.link(
                    _("Generation %d") % g,
                    "PersonList",
                    handles,
                    tooltip=_("Double-click to see people in generation %d") % g,
                )
                percent = (
                    glocale.format_string("%.2f", float(count) / 2 ** (g - 1) * 100)
                    + percent_sign
                )
                self.append_text(
                    # Translators: leave all/any {...} untranslated
                    ngettext(
                        " has {count_person} of {max_count_person} "
                        "individuals ({percent} complete)\n",
                        " has {count_person} of {max_count_person} "
                        "individuals ({percent} complete)\n",
                        2 ** (g - 1),
                    ).format(
                        count_person=count,
                        max_count_person=2 ** (g - 1),
                        percent=percent,
                    )
                )
        self.link(
            _("All generations"),
            "PersonList",
            all,
            tooltip=_("Double-click to see all generations"),
        )
        self.append_text(
            # Translators: leave all/any {...} untranslated
            ngettext(
                " have {number_of} individual\n",
                " have {number_of} individuals\n",
                len(all),
            ).format(number_of=len(all))
        )
        # Set to a fixed font
        if self.box_mode == "UTF":
            start, end = self.gui.buffer.get_bounds()
            self.gui.buffer.apply_tag_by_name("fixed", start, end)
        self.append_text("", scroll_to="begin")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Source Attribute class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .attribute import AttributeRoot
from .srcattrtype import SrcAttributeType
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Attribute for Source/Citation
#
# -------------------------------------------------------------------------
class SrcAttribute(AttributeRoot):
    """
    Provide a simple key/value pair for describing properties.
    Used to store descriptive information.
    """

    def __init__(self, source=None):
        """
        Create a new Attribute object, copying from the source if provided.
        """
        AttributeRoot.__init__(self, source)

        if source:
            self.type = SrcAttributeType(source.type)
            self.value = source.value
        else:
            self.type = SrcAttributeType()
            self.value = ""

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        return {
            "type": "object",
            "title": _("Attribute"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "private": {"type": "boolean", "title": _("Private")},
                "type": SrcAttributeType.get_schema(),
                "value": {"type": "string", "title": _("Value")},
            },
        }



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Brian G. Matherly
# Copyright (C) 2009       Peter G. Landgren
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Some independent constants/functions that can be safely imported
without any translation happening yet.  Do _not_ add imports that will
perform a translation on import, eg Gtk.
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
import platform
import sys
import os

# -------------------------------------------------------------------------
#
# Platforms
# Never test on LINUX, handle Linux in the else statement as default
#
# -------------------------------------------------------------------------
LINUX = ["Linux", "linux", "linux2"]
MACOS = ["Darwin", "darwin"]
WINDOWS = ["Windows", "win32"]

# -------------------------------------------------------------------------
#
# Platform determination functions
#
# -------------------------------------------------------------------------


def lin():
    """
    Return True if a linux system
    Note: Normally do as linux in else statement of a check !
    """
    if platform.system() in LINUX:
        return True
    return False


def mac():
    """
    Return True if a Macintosh system
    """
    if platform.system() in MACOS:
        return True
    return False


def win():
    """
    Return True if a windows system
    """
    if platform.system() in WINDOWS:
        return True
    return False


## The following functions do import gtk, but only when called. They
## should only be called after translation system has been
## initialized!


def is_quartz():
    """
    Tests to see if Python is currently running with gtk and
    windowing system is Mac OS-X's "quartz".
    """
    if mac():
        try:
            import gi

            gi.require_version("Gtk", "3.0")
            gi.require_version("Gdk", "3.0")
            from gi.repository import Gtk
            from gi.repository import Gdk
        except ImportError:
            return False
        return Gdk.Display.get_default().__class__.__name__.endswith("QuartzDisplay")
    return False


def has_display():
    """
    Tests to see if Python is currently running with gtk
    """
    # FIXME: currently, Gtk.init_check() requires all strings
    # in argv, and we might have unicode.
    temp, sys.argv = sys.argv, sys.argv[:1]
    try:
        import gi

        gi.require_version("Gtk", "3.0")
        gi.require_version("Gdk", "3.0")
        from gi.repository import Gtk
        from gi.repository import Gdk
    except ImportError:
        return False

    try:
        test = Gtk.init_check(temp) and Gdk.Display.get_default()
        sys.argv = temp
        return bool(test)
    except:
        sys.argv = temp
        return False


# A couple of places add menu accelerators using <alt>, which doesn't
# work with Gtk-quartz. <Meta> is the usually correct replacement, but
# in one case the key is a number, and <meta>number is used by Spaces
# (a mac feature), so we'll use control instead.


def mod_key():
    """
    Returns a string to pass to an accelerator map.
    """

    if is_quartz():
        return "<ctrl>"

    return "<alt>"


# Python2 on Windows munges environemnt variables to match the system
# code page. This breaks all manner of things and the workaround
# though a bit ugly, is encapsulated here. Use this to retrieve
# environment variables if there's any chance they might contain
# Unicode, and especially for paths.

# Shamelessly lifted from http://stackoverflow.com/questions/2608200/problems-with-umlauts-in-python-appdata-environvent-variable, answer 1.


def get_env_var(name, default=None):
    """
    Python2 on Windows can't directly read unicode values from
    environment variables. This routine does so using the native C
    wide-character function.
    """
    if not name or name not in os.environ:
        return default

    return os.environ[name]


def get_curr_dir():
    """
    In Python2 on Windows, os.getcwd() returns a string encoded with
    the current code page, which may not be able to correctly handle
    an arbitrary unicode character in a path. This function uses the
    native GetCurrentDirectory function to return a unicode cwd.
    """
    return os.getcwd()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2005 Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Exceptions generated by the Db package."""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import URL_WIKISTRING, URL_MANUAL_PAGE
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
URL_WIKI_START = '<a href="' + URL_WIKISTRING
URL_MANUAL_START = URL_WIKI_START + URL_MANUAL_PAGE + "_-_Manage_Family_Trees"
URL_BACKUP1_START = URL_WIKI_START + 'How_to_make_a_backup">'
URL_BACKUP2_START = URL_MANUAL_START + '#Backing_up_a_Family_Tree">'
URL_EXPORT_START = URL_MANUAL_START + '#Export_into_Gramps_formats">'


class DbException(Exception):
    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        return self.value


class DbWriteFailure(Exception):
    """
    Error used to indicate that a write to a database has failed.
    """

    def __init__(self, value, value2=""):
        Exception.__init__(self)
        self.value = value
        self.value2 = value2

    def __str__(self):
        return self.value

    def messages(self):
        return self.value, self.value2


class DbTransactionCancel(Exception):
    """
    Error used to indicate that a transaction needs to be canceled,
    for example becuase it is lengthy and the users requests so.
    """

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        return self.value


class DbVersionError(Exception):
    """
    Error used to report that a file could not be read because it is written
    in an unsupported version of the file format.
    """

    def __init__(self, tree_vers, min_vers, max_vers):
        Exception.__init__(self)
        self.tree_vers = tree_vers
        self.min_vers = min_vers
        self.max_vers = max_vers

    def __str__(self):
        return _(
            "The schema version is not supported by this version of "
            "Gramps.\n\n"
            "This Family Tree is schema version %(tree_vers)s, and this "
            "version of Gramps supports versions %(min_vers)s to "
            "%(max_vers)s\n\n"
            "Please upgrade to the corresponding version or use "
            "XML for porting data between different schema versions."
        ) % {
            "tree_vers": self.tree_vers,
            "min_vers": self.min_vers,
            "max_vers": self.max_vers,
        }


class DbPythonError(Exception):
    """
    Error used to report that a file could not be read because it is written
    in an unsupported version of the Python format.
    """

    def __init__(self, tree_vers, min_vers, max_vers):
        Exception.__init__(self)
        self.tree_vers = tree_vers
        self.min_vers = min_vers
        self.max_vers = max_vers

    def __str__(self):
        return _(
            "The Python version is not supported by this version of "
            "Gramps.\n\n"
            "This Family Tree is Python version %(tree_vers)s, and this "
            "version of Gramps supports versions %(min_vers)s to "
            "%(max_vers)s\n\n"
            "Please upgrade to the corresponding version or use "
            "XML for porting data between different Python versions."
        ) % {
            "tree_vers": self.tree_vers,
            "min_vers": self.min_vers,
            "max_vers": self.max_vers,
        }


class DbUpgradeRequiredError(Exception):
    """
    Error used to report that a database needs to be upgraded before it can be
    used.
    """

    def __init__(self, oldschema, newschema):
        Exception.__init__(self)
        self.oldschema = oldschema
        self.newschema = newschema

    def __str__(self):
        return _(
            "The Family Tree you are trying to load is in the schema "
            "version %(oldschema)s format. This version of Gramps uses "
            "schema version %(newschema)s. Therefore you cannot load this "
            "Family Tree without upgrading the schema version of the "
            "Family Tree.\n\n"
            "If you upgrade then you won't be able to use the previous "
            "version of Gramps, even if you subsequently "
            "%(wiki_manual_backup_html_start)sbackup%(html_end)s or "
            "%(wiki_manual_export_html_start)sexport%(html_end)s "
            "your upgraded Family Tree.\n\n"
            "Upgrading is a difficult task which could irretrievably "
            "corrupt your Family Tree if it is interrupted or fails.\n\n"
            "If you have not already made a backup of your Family Tree, "
            "then you should start your %(bold_start)sold%(bold_end)s "
            "version of Gramps and "
            "%(wiki_backup_html_start)smake a backup%(html_end)s "
            "of your Family Tree."
        ) % {
            "wiki_backup_html_start": URL_BACKUP1_START,
            "wiki_manual_backup_html_start": URL_BACKUP2_START,
            "wiki_manual_export_html_start": URL_EXPORT_START,
            "html_end": "</a>",
            "bold_start": "<b>",
            "bold_end": "</b>",
            "oldschema": self.oldschema,
            "newschema": self.newschema,
        }


class DbConnectionError(Exception):
    """
    Error used to report that a database connection failed.
    """

    def __init__(self, msg, settings_file):
        Exception.__init__(self)
        self.msg = msg
        self.settings_file = settings_file

    def __str__(self):
        return _(
            "Database connection failed.\n\n"
            "%(message)s\n"
            "Please check your connection settings file:\n"
            "%(settings_file)s"
        ) % {"message": self.msg, "settings_file": self.settings_file}


class DbSupportedError(Exception):
    """
    Error used to report that a database is no longer supported.
    """

    def __init__(self, msg):
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        return _(
            "The Family Tree you are trying to load is in the %(dbtype)s "
            "database, which is no longer supported.\nTherefore you "
            "cannot load this Family Tree without upgrading.\n\n"
            "If you upgrade then you won't be able to use the previous "
            "version of Gramps, even if you subsequently "
            "%(wiki_manual_backup_html_start)sbackup%(html_end)s or "
            "%(wiki_manual_export_html_start)sexport%(html_end)s "
            "your upgraded Family Tree.\n\n"
            "You are strongly advised to backup your Family Tree.\n\n"
            "If you have not already made a backup of your Family Tree, "
            "then you should start your previous version of Gramps and "
            "%(wiki_backup_html_start)smake a backup%(html_end)s "
            "of your Family Tree."
        ) % {
            "dbtype": self.msg,
            "wiki_manual_backup_html_start": URL_BACKUP2_START,
            "wiki_manual_export_html_start": URL_EXPORT_START,
            "wiki_backup_html_start": URL_BACKUP1_START,
            "html_end": "</a>",
        }


if __name__ == "__main__":
    """
    Call this from the CLI (in order to find the imported modules):
        cd /path/to/gramps; GRAMPS_RESOURCES=. ; PYTHONPATH=.
        python gramps/gen/db/exceptions.py
    """
    import sys

    print("\nDbVersionError:\n", DbVersionError("1.6.0", "1.5.0", "1.5.1"))
    print("\nDbUpgradeRequiredError:\n", DbUpgradeRequiredError("1.5.1", "1.6.0"))
    sys.exit(0)

    print("\nxxx:\n", xxx("4.8.30", "4.8.29"))



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007  Donald N. Allingham, Martin Hawlisch
# Copyright (C) 2009 Douglas S. Blank
# Copyright (C) 2009-2016 Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

# the print settings to remember between print sessions
PRINT_SETTINGS = None


# ------------------------------------------------------------------------
#
# CairoPrintSave class
#
# ------------------------------------------------------------------------
class CairoPrintSave:
    """
    Act as an abstract document that can render onto a cairo context.

    It can render the model onto cairo context pages, according to the received
    page style.

    """

    def __init__(self, widthpx, heightpx, drawfunc, parent):
        """
        This class provides the things needed so as to dump a cairo drawing on
        a context to output
        """
        # We try to adapt the cairo dump to an A4 page.
        widthpx += 30
        heightpx += 30
        self.scale = 100.0
        self.widthpx = widthpx
        self.heightpx = heightpx
        self.drawfunc = drawfunc
        self.parent = parent
        self.preview = None

    def run(self):
        """
        Create the physical output from the meta document.

        """
        global PRINT_SETTINGS

        # set up a print operation
        operation = Gtk.PrintOperation()
        operation.connect("draw_page", self.on_draw_page)
        operation.connect("preview", self.on_preview)
        operation.set_n_pages(1)
        page_setup = Gtk.PageSetup()
        if PRINT_SETTINGS is None:
            PRINT_SETTINGS = Gtk.PrintSettings()
        page_setup = Gtk.print_run_page_setup_dialog(None, page_setup, PRINT_SETTINGS)
        paper_size_used = page_setup.get_paper_size()
        if self.widthpx > self.heightpx:
            szw = self.widthpx
            szh = self.heightpx
        else:
            szh = self.widthpx
            szw = self.heightpx
        height_used = paper_size_used.get_height(Gtk.Unit.POINTS)
        width_used = paper_size_used.get_width(Gtk.Unit.POINTS)
        coefx = szw / height_used  # width and height depends on the selected
        # page (A4, A3, ...)
        coefy = szh / width_used
        if coefx < coefy:
            self.scale = 100.0 / coefy
        else:
            self.scale = 100.0 / coefx
        PRINT_SETTINGS.set_scale(self.scale)
        if self.widthpx > self.heightpx:
            page_setup.set_orientation(Gtk.PageOrientation.LANDSCAPE)
        else:
            page_setup.set_orientation(Gtk.PageOrientation.PORTRAIT)
        operation.set_default_page_setup(page_setup)

        if PRINT_SETTINGS is not None:
            operation.set_print_settings(PRINT_SETTINGS)

        # run print dialog
        while True:
            self.preview = None
            res = operation.run(Gtk.PrintOperationAction.PRINT_DIALOG, None)
            if self.preview is None:  # cancel or print
                break
            # set up printing again; can't reuse PrintOperation?
            operation = Gtk.PrintOperation()
            operation.set_default_page_setup(page_setup)
            operation.connect("draw_page", self.on_draw_page)
            operation.connect("preview", self.on_preview)
            # set print settings if it was stored previously
            if PRINT_SETTINGS is not None:
                operation.set_print_settings(PRINT_SETTINGS)

        # store print settings if printing was successful
        if res == Gtk.PrintOperationResult.APPLY:
            PRINT_SETTINGS = operation.get_print_settings()

    def on_draw_page(self, operation, context, page_nr):
        """Draw a page on a Cairo context."""
        dummy_op = operation
        dummy_ct = context
        dummy_pg = page_nr
        ctx = context.get_cairo_context()
        self.drawfunc(self.parent, ctx)

    def on_preview(self, operation, preview, context, parent):
        """Implement custom print preview functionality.
        We provide a dummy holder here, because on_preview crashes if no
        default application is set with gir 3.3.2 (typically evince not
        installed)!
        """
        dummy_op = operation
        dummy_ct = context
        dummy_pr = preview
        dummy_pa = parent
        operation.run(Gtk.PrintOperationAction.PREVIEW, None)
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2021  Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
An override to allow resizable columns
"""
import logging

from gi.repository import Gtk
from gramps.gen.config import config

_LOG = logging.getLogger(".persistent")

# -------------------------------------------------------------------------
#
# PersistentTreeView class
#
# -------------------------------------------------------------------------

__all__ = ["PersistentTreeView"]


class PersistentTreeView(Gtk.TreeView):
    """
    TreeView that has resizable columns
    """

    __gtype_name__ = "PersistentTreeView"

    def __init__(self, uistate=None, config_name=None):
        """
        Create a TreeView widgets with column size saving
        """
        Gtk.TreeView.__init__(self)
        self.config_name = "undefined"
        self.connect("destroy", self.save_column_info)
        self.uistate = None
        if uistate:
            self.set_uistate(uistate)
        if config_name:
            self.set_config_name(config_name)

    def set_uistate(self, uistate):
        """
        parameter: uistate: The associated uistate
        This parameter is used to connect some signals from this class.
        """
        if not self.uistate and uistate:
            _LOG.debug("connect signal font-changed")
            uistate.connect("font-changed", self.restore_column_size)
            self.uistate = uistate

    def set_config_name(self, name):
        """
        parameter: name: The associated config name string for the treeview
        This parameter must be unique as it allow the tab columns size saving
        """
        # Here, we can have:
        # name = gramps.gui.editors.displaytabs.personeventembedlist
        # The following line return only the last part of the string.
        # personeventembedlist in this case
        last = name.split(".")[-1]
        _LOG.debug("set persistent name : %s" % last)
        self.config_name = "spacing.%s" % last
        if not config.is_set(self.config_name):
            _LOG.debug("registering %s" % self.config_name)
            config.register(self.config_name, [])

    def save_column_info(self, tree=None):
        """
        Save the columns width
        """
        if self.config_name == "undefined":
            return
        newsize = self.get_columns_size()
        if 0 not in newsize:
            # Don't save the values if one column size is null.
            config.set(self.config_name, newsize)
            _LOG.debug("save persistent : %s = %s" % (self.config_name, newsize))
        return

    def get_columns_size(self):
        """
        Get all the columns size
        """
        columns = self.get_columns()
        newsize = []
        nbc = 0
        context = Gtk.Label().get_pango_context()
        font_desc = context.get_font_description()
        char_width = 1.2 * (font_desc.get_size() / 1000)
        for column in columns:
            if nbc < len(columns) - 1 or len(columns) == 1:
                # Don't save the last column size, it's wrong
                # except if we have only one column (i.e. EditRule)
                child = column.get_widget()
                if isinstance(child, Gtk.Image):
                    # Don't resize the icons
                    size = 2
                else:
                    size = column.get_width() / char_width
                    size = 2 if size < 2 else size
                newsize.append(size)
            nbc += 1
        return newsize

    def restore_column_size(self):
        """
        restore the columns width
        """
        if self.config_name == "undefined":
            return
        size = config.get(self.config_name)
        if len(size) == 0:
            _LOG.debug(
                "restore for the first time : %s = %s" % (self.config_name, size)
            )
            return
        _LOG.debug("restore persistent : %s = %s" % (self.config_name, size))
        context = Gtk.Label().get_pango_context()
        font_desc = context.get_font_description()
        char_width = 1.2 * (font_desc.get_size() / 1000)
        nbc = 0
        columns = self.get_columns()
        for column in columns:
            if nbc < len(size):
                column.set_fixed_width(size[nbc] * char_width)
            nbc += 1



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009-2013  Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Url class for Gramps.
"""

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
from warnings import warn
from urllib.parse import urlparse

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .secondaryobj import SecondaryObject
from .privacybase import PrivacyBase
from .urltype import UrlType
from .const import IDENTICAL, EQUAL, DIFFERENT
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Url for Person/Place/Repository
#
# -------------------------------------------------------------------------
class Url(SecondaryObject, PrivacyBase):
    """
    Contains information related to internet Uniform Resource Locators,
    allowing gramps to store information about internet resources.
    """

    def __init__(self, source=None):
        """Create a new URL instance, copying from the source if present."""
        PrivacyBase.__init__(self, source)
        if source:
            self.path = source.path
            self.desc = source.desc
            self.type = UrlType(source.type)
        else:
            self.path = ""
            self.desc = ""
            self.type = UrlType()

    def serialize(self):
        return (self.private, self.path, self.desc, self.type.serialize())

    def unserialize(self, data):
        (self.private, self.path, self.desc, type_value) = data
        self.type.unserialize(type_value)
        return self

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        return {
            "type": "object",
            "title": _("Url"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "private": {"type": "boolean", "title": _("Private")},
                "path": {"type": "string", "title": _("Path")},
                "desc": {"type": "string", "title": _("Description")},
                "type": UrlType.get_schema(),
            },
        }

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.path, self.desc]

    def is_equivalent(self, other):
        """
        Return if this url is equivalent, that is agrees in type, full path
        name and description, to other.

        :param other: The url to compare this one to.
        :type other: Url
        :returns: Constant indicating degree of equivalence.
        :rtype: int
        """
        if (
            self.type != other.type
            or self.get_full_path() != other.get_full_path()
            or self.desc != other.desc
        ):
            return DIFFERENT
        else:
            if self.get_privacy() != other.get_privacy():
                return EQUAL
            else:
                return IDENTICAL

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this url.

        :param acquisition: The url to merge with the present url.
        :type acquisition: Url
        """
        self._merge_privacy(acquisition)

    def set_path(self, path):
        """Set the URL path."""
        self.path = path

    def get_path(self):
        """Return the URL path."""
        return self.path

    def set_description(self, description):
        """Set the description of the URL."""
        self.desc = description

    def get_description(self):
        """Return the description of the URL."""
        return self.desc

    def set_type(self, the_type):
        """
        :param the_type: descriptive type of the Url
        :type the_type: str
        """
        self.type.set(the_type)

    def get_type(self):
        """
        :returns: the descriptive type of the Url
        :rtype: str
        """
        return self.type

    def are_equal(self, other):
        """Deprecated - use :meth:`~.SecondaryObject.is_equal` instead."""

        warn("Use is_equal instead of are_equal", DeprecationWarning, 2)
        return self.is_equal(other)

    def parse_path(self):
        """
        Returns a 6 tuple-based object with the following items:

        ======== ===   =================================
        Property Pos   Meaning
        ======== ===   =================================
        scheme   0     URL scheme specifier
        netloc   1     Network location part
        path     2     Hierarchical path
        params   3     Parameters for last path element
        query    4     Query component
        fragment 5     Fragment identifier
        ======== ===   =================================
        """
        return urlparse(self.path)

    def get_full_path(self):
        """
        Returns a full url, complete with scheme, even if missing from path.
        """
        if self.type == UrlType.EMAIL and not self.path.startswith("mailto:"):
            return "mailto:" + self.path
        elif self.type == UrlType.WEB_FTP and not self.path.startswith("ftp://"):
            return "ftp://" + self.path
        elif self.parse_path().scheme == "":
            return "http://" + self.path
        else:
            return self.path



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.lib import Place, PlaceRef
from gramps.gen.errors import WindowActiveError
from ...dialog import ErrorDialog
from ...ddtargets import DdTargets
from .placerefmodel import PlaceRefModel
from .embeddedlist import EmbeddedList, TEXT_COL
from ...selectors import SelectorFactory
from ...dbguielement import DbGUIElement


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class PlaceRefEmbedList(DbGUIElement, EmbeddedList):
    _HANDLE_COL = 4
    _DND_TYPE = DdTargets.PLACEREF
    _DND_EXTRA = DdTargets.PLACE_LINK

    # index = column in model. Value =
    #  (name, sortcol in model, width, markup/text, weigth_col
    _column_names = [
        (_("ID"), 0, 75, TEXT_COL, -1, None),
        (_("Name"), 1, 250, TEXT_COL, -1, None),
        (_("Type"), 2, 100, TEXT_COL, -1, None),
        (_("Date"), 3, 150, TEXT_COL, -1, None),
    ]

    def __init__(self, dbstate, uistate, track, data, config_key, handle, callback):
        self.data = data
        self.handle = handle
        self.callback = callback
        DbGUIElement.__init__(self, dbstate.db)
        EmbeddedList.__init__(
            self,
            dbstate,
            uistate,
            track,
            _("Enclosed By"),
            PlaceRefModel,
            config_key,
            share_button=True,
            move_buttons=True,
        )

    def _connect_db_signals(self):
        """
        called on init of DbGUIElement, connect to db as required.
        """
        # note: place-rebuild closes the editors, so no need to connect to it
        self.callman.register_callbacks(
            {
                "place-update": self.place_change,  # change to place we track
                "place-delete": self.place_delete,  # delete of place we track
            }
        )
        self.callman.connect_all(keys=["place"])

    def place_change(self, *obj):
        """
        Callback method called when a tracked place changes (description
        changes...)
        """
        self.rebuild()

    def place_delete(self, hndls):
        """
        Callback method called when a tracked place is deleted.
        There are two possibilities:
        * a tracked non-workgroup place is deleted, just rebuilding the view
            will correct this.
        * a workgroup place is deleted. The place must be removed from the
            obj so that no inconsistent data is shown.
        """
        for handle in hndls:
            ref_list = [pref.ref for pref in self.data]
            indexlist = []
            last = -1
            while True:
                try:
                    last = ref_list.index(handle, last + 1)
                    indexlist.append(last)
                except ValueError:
                    break
            # remove the deleted workgroup places from the object
            for index in reversed(indexlist):
                del self.data[index]
        # now rebuild the display tab
        self.rebuild()

    def get_data(self):
        self.callman.register_handles({"place": [pref.ref for pref in self.data]})
        return self.data

    def column_order(self):
        return ((1, 0), (1, 1), (1, 2), (1, 3))

    def get_skip_list(self, handle):
        todo = [handle]
        skip = [handle]
        while todo:
            handle = todo.pop()
            for child in self.dbstate.db.find_backlink_handles(handle, ["Place"]):
                if child[1] not in skip:
                    todo.append(child[1])
                    skip.append(child[1])
        return skip

    def add_button_clicked(self, obj):
        ref = PlaceRef()
        place = Place()
        try:
            from .. import EditPlaceRef

            EditPlaceRef(
                self.dbstate, self.uistate, self.track, place, ref, self.add_callback
            )
        except WindowActiveError:
            pass

    def add_callback(self, ref, place):
        ref.ref = place.handle
        data = self.get_data()
        data.append(ref)
        self.rebuild()
        GLib.idle_add(self.tree.scroll_to_cell, len(data) - 1)

    def share_button_clicked(self, obj):
        SelectPlace = SelectorFactory("Place")

        sel = SelectPlace(
            self.dbstate, self.uistate, self.track, skip=self.get_skip_list(self.handle)
        )
        place = sel.run()
        if place:
            ref = PlaceRef()
            try:
                from .. import EditPlaceRef

                EditPlaceRef(
                    self.dbstate,
                    self.uistate,
                    self.track,
                    place,
                    ref,
                    self.add_callback,
                )
            except WindowActiveError:
                pass

    def edit_button_clicked(self, obj):
        ref = self.get_selected()
        if ref:
            place = self.dbstate.db.get_place_from_handle(ref.ref)
            try:
                from .. import EditPlaceRef

                EditPlaceRef(
                    self.dbstate,
                    self.uistate,
                    self.track,
                    place,
                    ref,
                    self.edit_callback,
                )
            except WindowActiveError:
                pass

    def edit_callback(self, ref, place):
        self.rebuild()

    def post_rebuild(self, prebuildpath):
        self.callback()

    def handle_extra_type(self, objtype, obj):
        if obj in self.get_skip_list(self.handle):
            ErrorDialog(
                _("Place cycle detected"),
                _("The place you are adding is already enclosed by " "this place"),
                parent=self.uistate.window,
            )
            return
        place = self.dbstate.db.get_place_from_handle(obj)
        placeref = PlaceRef()
        try:
            from .. import EditPlaceRef

            EditPlaceRef(
                self.dbstate,
                self.uistate,
                self.track,
                place,
                placeref,
                self.add_callback,
            )
        except WindowActiveError:
            pass



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2015       Gerald Kunzmann <gerald@gkunzmann.de>
# Copyright (C) 2013-2016  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Reports/Text Reports/Family Group Report"""

# ------------------------------------------------------------------------
#
# Python Library
#
# ------------------------------------------------------------------------
from functools import partial

# ------------------------------------------------------------------------
#
# Gramps
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.lib import EventRoleType, EventType, NoteType, Person
from gramps.gen.plug.menu import BooleanOption, FamilyOption, FilterOption
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.plug.docgen import (
    IndexMark,
    FontStyle,
    ParagraphStyle,
    TableStyle,
    TableCellStyle,
    FONT_SANS_SERIF,
    FONT_SERIF,
    INDEX_TYPE_TOC,
    PARA_ALIGN_CENTER,
)
from gramps.gen.display.place import displayer as _pd
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.errors import ReportError
from gramps.gen.utils.db import family_name


# ------------------------------------------------------------------------
#
# FamilyGroup
#
# ------------------------------------------------------------------------
class FamilyGroup(Report):
    """Family Group Report"""

    def __init__(self, database, options, user):
        """
        Create the FamilyGroup object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - a gen.user.User() instance

        This report needs the following parameters (class variables)
        that come in the options class.

        filter    - Filter to be applied to the families of the database.
                    The option class carries its number, and the function
                    returning the list of filters.
        incattrs      - Whether to include attributes
        name_format   - Preferred format to display names
        incl_private  - Whether to include private data
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        """
        Report.__init__(self, database, options, user)
        self._user = user
        menu = options.menu

        self.set_locale(menu.get_option_by_name("trans").get_value())
        self._ = self._locale.translation.sgettext  # needed for English

        stdoptions.run_date_format_option(self, menu)

        stdoptions.run_private_data_option(self, menu)
        stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)
        self.db = self.database

        self.filter = menu.get_option_by_name("filter").get_filter()

        get_option_by_name = menu.get_option_by_name
        get_value = lambda name: get_option_by_name(name).get_value()
        self.gramps_ids = get_value("inc_id")
        self.recursive = get_value("recursive")
        self.missing_info = get_value("missinginfo")
        self.generations = get_value("generations")
        self.inc_fam_notes = get_value("incFamNotes")
        self.inc_par_events = get_value("incParEvents")
        self.inc_par_addr = get_value("incParAddr")
        self.inc_par_notes = get_value("incParNotes")
        self.inc_par_names = get_value("incParNames")
        self.inc_par_mar = get_value("incParMar")
        self.inc_rel_dates = get_value("incRelDates")
        self.inc_chi_mar = get_value("incChiMar")
        self.include_attrs = get_value("incattrs")

        stdoptions.run_name_format_option(self, menu)

        self.place_format = menu.get_option_by_name("place_format").get_value()

    def dump_parent_event(self, name, event):
        place = ""
        date = ""
        descr = ""
        if event:
            date = self._get_date(event.get_date_object())
            place = _pd.display_event(self.db, event, self.place_format)
            if place is None:
                place = ""
            descr = event.get_description()

            if self.include_attrs:
                for attr in event.get_attribute_list():
                    if descr:
                        # Translators: needed for Arabic, ignore otherwise
                        descr += self._("; ")
                    attr_type = self._get_type(attr.get_type())
                    # Translators: needed for French, ignore otherwise
                    descr += self._("%(str1)s: %(str2)s") % {
                        "str1": self._(attr_type),
                        "str2": attr.get_value(),
                    }

        self.doc.start_row()
        self.doc.start_cell("FGR-TextContents")
        self.doc.start_paragraph("FGR-Normal")
        self.doc.write_text(name)
        self.doc.end_paragraph()
        self.doc.end_cell()

        if descr:
            self.doc.start_cell("FGR-TextContentsEnd", 2)
            self.doc.start_paragraph("FGR-Normal")
            self.doc.write_text(descr)
            self.doc.end_paragraph()
            self.doc.end_cell()
            self.doc.end_row()

            if date or place:
                self.doc.start_row()
                self.doc.start_cell("FGR-TextContents")
                self.doc.start_paragraph("FGR-Normal")
                self.doc.end_paragraph()
                self.doc.end_cell()

        if (date or place) or not descr:
            self.doc.start_cell("FGR-TextContents")
            self.doc.start_paragraph("FGR-Normal")
            self.doc.write_text(date)
            self.doc.end_paragraph()
            self.doc.end_cell()
            self.doc.start_cell("FGR-TextContentsEnd")
            self.doc.start_paragraph("FGR-Normal")
            self.doc.write_text(place)
            self.doc.end_paragraph()
            self.doc.end_cell()
            self.doc.end_row()

    def dump_parent_parents(self, person):
        family_handle = person.get_main_parents_family_handle()
        father_name = ""
        mother_name = ""
        if family_handle:
            family = self.db.get_family_from_handle(family_handle)
            father_handle = family.get_father_handle()
            if father_handle:
                father = self.db.get_person_from_handle(father_handle)
                father_name = self._name_display.display(father)
                if self.gramps_ids:
                    gid = father.get_gramps_id()
                    if gid:
                        father_name += " (%s)" % gid
                if self.inc_rel_dates:
                    birth_ref = father.get_birth_ref()
                    birth = "  "
                    if birth_ref:
                        event = self.db.get_event_from_handle(birth_ref.ref)
                        birth = self._get_date(event.get_date_object())
                    death_ref = father.get_death_ref()
                    death = "  "
                    if death_ref:
                        event = self.db.get_event_from_handle(death_ref.ref)
                        death = self._get_date(event.get_date_object())
                    if birth_ref or death_ref:
                        father_name += " (%s - %s)" % (birth, death)
            mother_handle = family.get_mother_handle()
            if mother_handle:
                mother = self.db.get_person_from_handle(mother_handle)
                mother_name = self._name_display.display(mother)
                if self.gramps_ids:
                    gid = mother.get_gramps_id()
                    if gid:
                        mother_name += " (%s)" % gid
                if self.inc_rel_dates:
                    birth_ref = mother.get_birth_ref()
                    birth = "  "
                    if birth_ref:
                        event = self.db.get_event_from_handle(birth_ref.ref)
                        birth = self._get_date(event.get_date_object())
                    death_ref = mother.get_death_ref()
                    death = "  "
                    if death_ref:
                        event = self.db.get_event_from_handle(death_ref.ref)
                        death = self._get_date(event.get_date_object())
                    if birth_ref or death_ref:
                        mother_name += " (%s - %s)" % (birth, death)

        if father_name != "":
            self.doc.start_row()
            self.doc.start_cell("FGR-TextContents")
            self.doc.start_paragraph("FGR-Normal")
            self.doc.write_text(self._("Father"))
            self.doc.end_paragraph()
            self.doc.end_cell()
            self.doc.start_cell("FGR-TextContentsEnd", 2)
            self.doc.start_paragraph("FGR-Normal")
            mark = utils.get_person_mark(self.db, father)
            self.doc.write_text(father_name, mark)
            self.doc.end_paragraph()
            self.doc.end_cell()
            self.doc.end_row()
        elif self.missing_info:
            self.dump_parent_line(self._("Father"), "")

        if mother_name != "":
            self.doc.start_row()
            self.doc.start_cell("FGR-TextContents")
            self.doc.start_paragraph("FGR-Normal")
            self.doc.write_text(self._("Mother"))
            self.doc.end_paragraph()
            self.doc.end_cell()
            self.doc.start_cell("FGR-TextContentsEnd", 2)
            self.doc.start_paragraph("FGR-Normal")
            mark = utils.get_person_mark(self.db, mother)
            self.doc.write_text(mother_name, mark)
            self.doc.end_paragraph()
            self.doc.end_cell()
            self.doc.end_row()
        elif self.missing_info:
            self.dump_parent_line(self._("Mother"), "")

    def dump_parent_line(self, name, text):
        self.doc.start_row()
        self.doc.start_cell("FGR-TextContents")
        self.doc.start_paragraph("FGR-Normal")
        self.doc.write_text(name)
        self.doc.end_paragraph()
        self.doc.end_cell()
        self.doc.start_cell("FGR-TextContentsEnd", 2)
        self.doc.start_paragraph("FGR-Normal")
        self.doc.write_text(text)
        self.doc.end_paragraph()
        self.doc.end_cell()
        self.doc.end_row()

    def dump_parent_noteline(self, name, note):
        self.doc.start_row()
        self.doc.start_cell("FGR-TextContents")
        self.doc.start_paragraph("FGR-Normal")
        self.doc.write_text(name)
        self.doc.end_paragraph()
        self.doc.end_cell()
        self.doc.start_cell("FGR-TextContentsEnd", 2)
        self.doc.write_styled_note(
            note.get_styledtext(),
            note.get_format(),
            "FGR-Note",
            contains_html=(note.get_type() == NoteType.HTML_CODE),
        )
        self.doc.end_cell()
        self.doc.end_row()

    def dump_parent(self, table_name, person_handle):
        if not person_handle and not self.missing_info:
            return
        elif not person_handle:
            person = Person()
        else:
            person = self.db.get_person_from_handle(person_handle)
        name = self._name_display.display(person)

        self.doc.start_table(table_name, "FGR-ParentTable")
        self.doc.start_row()
        self.doc.start_cell("FGR-ParentHead", 3)
        self.doc.start_paragraph("FGR-ParentName")
        mark = utils.get_person_mark(self.db, person)
        self.doc.write_text(name, mark)
        if self.gramps_ids:
            gid = person.get_gramps_id()
            if gid:
                self.doc.write_text(" (%s)" % gid)
        self.doc.end_paragraph()
        self.doc.end_cell()
        self.doc.end_row()

        birth_ref = person.get_birth_ref()
        birth = None
        ev_name = self._("Birth")
        if birth_ref:
            birth = self.db.get_event_from_handle(birth_ref.ref)
        if birth or self.missing_info:
            self.dump_parent_event(ev_name, birth)

        death_ref = person.get_death_ref()
        death = None
        ev_name = self._("Death")
        if death_ref:
            death = self.db.get_event_from_handle(death_ref.ref)
        if death or self.missing_info:
            self.dump_parent_event(ev_name, death)

        self.dump_parent_parents(person)

        if self.inc_par_events:
            for event_ref in person.get_primary_event_ref_list():
                if event_ref != birth_ref and event_ref != death_ref:
                    event = self.db.get_event_from_handle(event_ref.ref)
                    event_type = self._get_type(event.get_type())
                    self.dump_parent_event(self._(event_type), event)

        if self.inc_par_addr:
            addrlist = person.get_address_list()[:]
            for addr in addrlist:
                location = utils.get_address_str(addr)
                date = self._get_date(addr.get_date_object())

                self.doc.start_row()
                self.doc.start_cell("FGR-TextContents")
                self.doc.start_paragraph("FGR-Normal")
                self.doc.write_text(self._("Address"))
                self.doc.end_paragraph()
                self.doc.end_cell()
                self.doc.start_cell("FGR-TextContents")
                self.doc.start_paragraph("FGR-Normal")
                self.doc.write_text(date)
                self.doc.end_paragraph()
                self.doc.end_cell()
                self.doc.start_cell("FGR-TextContentsEnd")
                self.doc.start_paragraph("FGR-Normal")
                self.doc.write_text(location)
                self.doc.end_paragraph()
                self.doc.end_cell()
                self.doc.end_row()

        if self.inc_par_notes:
            for notehandle in person.get_note_list():
                note = self.db.get_note_from_handle(notehandle)
                self.dump_parent_noteline(self._("Note"), note)

        if self.include_attrs:
            for attr in person.get_attribute_list():
                attr_type = self._get_type(attr.get_type())
                self.dump_parent_line(self._(attr_type), attr.get_value())

        if self.inc_par_names:
            for alt_name in person.get_alternate_names():
                name_type = self._get_type(alt_name.get_type())
                name = self._name_display.display_name(alt_name)
                self.dump_parent_line(self._(name_type), name)

        self.doc.end_table()

    def dump_marriage(self, family):
        if not family:
            return

        mrg = None
        family_list = family.get_event_ref_list()
        for event_ref in family_list:
            if event_ref:
                event = self.db.get_event_from_handle(event_ref.ref)
                if event.get_type() == EventType.MARRIAGE and (
                    event_ref.get_role() == EventRoleType.FAMILY
                    or event_ref.get_role() == EventRoleType.PRIMARY
                ):
                    mrg = event
                    break

        if len(family_list) > 0 or self.missing_info or self.include_attrs:
            self.doc.start_table("MarriageInfo", "FGR-ParentTable")
            self.doc.start_row()
            self.doc.start_cell("FGR-ParentHead", 3)
            self.doc.start_paragraph("FGR-ParentName")
            relationship_type = self._get_type(family.get_relationship())
            header = self._("Relationship: %s") % self._(relationship_type)
            if self.gramps_ids:
                header += " (%s)" % family.get_gramps_id()
            self.doc.write_text(header)
            self.doc.end_paragraph()
            self.doc.end_cell()
            self.doc.end_row()

            self.dump_parent_event(self._("Marriage"), mrg)

            for event_ref in family_list:
                if event_ref:
                    event = self.db.get_event_from_handle(event_ref.ref)
                    if event.get_type() != EventType.MARRIAGE:
                        event_type = self._get_type(event.get_type())
                        self.dump_parent_event(self._(event_type), event)

            if self.include_attrs:
                for attr in family.get_attribute_list():
                    attr_type = self._get_type(attr.get_type())
                    self.dump_parent_line(self._(attr_type), attr.get_value())

            if self.inc_fam_notes:
                for notehandle in family.get_note_list():
                    note = self.database.get_note_from_handle(notehandle)
                    self.dump_parent_noteline(self._("Note"), note)

            self.doc.end_table()

    def dump_child_event(self, text, name, event):
        date = ""
        place = ""
        if event:
            date = self._get_date(event.get_date_object())
            place_handle = event.get_place_handle()
            if place_handle:
                place = _pd.display_event(self.db, event, self.place_format)
                if place is None:
                    place = ""

        self.doc.start_row()
        self.doc.start_cell(text)
        self.doc.start_paragraph("FGR-Normal")
        self.doc.end_paragraph()
        self.doc.end_cell()
        self.doc.start_cell("FGR-TextContents")
        self.doc.start_paragraph("FGR-Normal")
        self.doc.write_text(name)
        self.doc.end_paragraph()
        self.doc.end_cell()
        self.doc.start_cell("FGR-TextContents")
        self.doc.start_paragraph("FGR-Normal")
        self.doc.write_text(date)
        self.doc.end_paragraph()
        self.doc.end_cell()
        self.doc.start_cell("FGR-TextContentsEnd")
        self.doc.start_paragraph("FGR-Normal")
        self.doc.write_text(place)
        self.doc.end_paragraph()
        self.doc.end_cell()
        self.doc.end_row()

    def dump_child(self, index, person_handle):
        person = self.db.get_person_from_handle(person_handle)
        families = len(person.get_family_handle_list())
        birth_ref = person.get_birth_ref()
        if birth_ref:
            birth = self.db.get_event_from_handle(birth_ref.ref)
        else:
            birth = None
        death_ref = person.get_death_ref()
        if death_ref:
            death = self.db.get_event_from_handle(death_ref.ref)
        else:
            death = None

        spouse_count = 0
        if self.inc_chi_mar:
            for family_handle in person.get_family_handle_list():
                family = self.db.get_family_from_handle(family_handle)
                spouse_id = None
                if person_handle == family.get_father_handle():
                    spouse_id = family.get_mother_handle()
                else:
                    spouse_id = family.get_father_handle()
                if spouse_id:
                    spouse_count += 1

        self.doc.start_row()
        if (
            spouse_count != 0
            or self.missing_info
            or death is not None
            or birth is not None
        ):
            self.doc.start_cell("FGR-TextChild1")
        else:
            self.doc.start_cell("FGR-TextChild2")
        self.doc.start_paragraph("FGR-ChildText")
        index_str = "%d" % index
        if person.get_gender() == Person.MALE:
            self.doc.write_text(index_str + self._("M", "acronym for male"))
        elif person.get_gender() == Person.FEMALE:
            self.doc.write_text(index_str + self._("F", "acronym for female"))
        elif person.get_gender() == Person.OTHER:
            self.doc.write_text(index_str + self._("X", "acronym for other"))
        else:
            self.doc.write_text(self._("%dU", "acronym for unknown") % index)
        self.doc.end_paragraph()
        self.doc.end_cell()

        name = self._name_display.display(person)
        mark = utils.get_person_mark(self.db, person)
        self.doc.start_cell("FGR-ChildName", 3)
        self.doc.start_paragraph("FGR-ChildText")
        self.doc.write_text(name, mark)
        if self.gramps_ids:
            self.doc.write_text(" (%s)" % person.get_gramps_id())
        self.doc.end_paragraph()
        self.doc.end_cell()
        self.doc.end_row()

        if self.missing_info or birth is not None:
            if spouse_count != 0 or self.missing_info or death is not None:
                self.dump_child_event("FGR-TextChild1", self._("Birth"), birth)
            else:
                self.dump_child_event("FGR-TextChild2", self._("Birth"), birth)

        if self.missing_info or death is not None:
            if spouse_count == 0 or not self.inc_chi_mar:
                self.dump_child_event("FGR-TextChild2", self._("Death"), death)
            else:
                self.dump_child_event("FGR-TextChild1", self._("Death"), death)

        if self.inc_chi_mar:
            index = 0
            for family_handle in person.get_family_handle_list():
                mrg = None
                index += 1
                family = self.db.get_family_from_handle(family_handle)

                for event_ref in family.get_event_ref_list():
                    if event_ref:
                        event = self.db.get_event_from_handle(event_ref.ref)
                        if event.type == EventType.MARRIAGE:
                            mrg = event
                            break

                spouse_id = None

                if person_handle == family.get_father_handle():
                    spouse_id = family.get_mother_handle()
                else:
                    spouse_id = family.get_father_handle()

                if spouse_id:
                    self.doc.start_row()
                    if mrg or index != families:
                        self.doc.start_cell("FGR-TextChild1")
                    else:
                        self.doc.start_cell("FGR-TextChild2")
                    self.doc.start_paragraph("FGR-Normal")
                    self.doc.end_paragraph()
                    self.doc.end_cell()
                    self.doc.start_cell("FGR-TextContents")
                    self.doc.start_paragraph("FGR-Normal")
                    self.doc.write_text(self._("Spouse"))
                    self.doc.end_paragraph()
                    self.doc.end_cell()
                    self.doc.start_cell("FGR-TextContentsEnd", 2)
                    self.doc.start_paragraph("FGR-Normal")

                    spouse = self.db.get_person_from_handle(spouse_id)
                    spouse_name = self._name_display.display(spouse)
                    if self.gramps_ids:
                        gid = spouse.get_gramps_id()
                        if gid:
                            spouse_name += " (%s)" % gid
                    if self.inc_rel_dates:
                        birth = "  "
                        birth_ref = spouse.get_birth_ref()
                        if birth_ref:
                            event = self.db.get_event_from_handle(birth_ref.ref)
                            birth = self._get_date(event.get_date_object())
                        death = "  "
                        death_ref = spouse.get_death_ref()
                        if death_ref:
                            event = self.db.get_event_from_handle(death_ref.ref)
                            death = self._get_date(event.get_date_object())
                        if birth_ref or death_ref:
                            spouse_name += " (%s - %s)" % (birth, death)
                    if self.gramps_ids:
                        fid = family.get_gramps_id()
                        if fid:
                            spouse_name += " (%s)" % fid
                    mark = utils.get_person_mark(self.db, spouse)
                    self.doc.write_text(spouse_name, mark)
                    self.doc.end_paragraph()
                    self.doc.end_cell()
                    self.doc.end_row()

                if mrg:
                    ev_name = self._("Marriage")
                    if index == families:
                        self.dump_child_event("FGR-TextChild2", ev_name, mrg)
                    else:
                        self.dump_child_event("FGR-TextChild1", ev_name, mrg)

    def dump_family(self, family_handle, generation):
        family = self.db.get_family_from_handle(family_handle)
        family_toc_name = family_name(family, self.db)

        self.doc.start_paragraph("FGR-Title")
        if self.recursive and self.generations:
            title = self._("Family Group Report - Generation %d") % generation
        else:
            title = self._("Family Group Report")
        mark = IndexMark(title, INDEX_TYPE_TOC, 1)
        self.doc.write_text(title, mark)
        mark = IndexMark(family_toc_name, INDEX_TYPE_TOC, 2)
        self.doc.write_text("", mark)
        self.doc.end_paragraph()

        self.dump_parent("Parent1", family.get_father_handle())
        self.doc.start_paragraph("FGR-blank")
        self.doc.end_paragraph()

        if self.inc_par_mar:
            self.dump_marriage(family)
            self.doc.start_paragraph("FGR-blank")
            self.doc.end_paragraph()

        self.dump_parent("Parent2", family.get_mother_handle())

        length = len(family.get_child_ref_list())
        if length > 0:
            self.doc.start_paragraph("FGR-blank")
            self.doc.end_paragraph()
            self.doc.start_table("FGR-Children", "FGR-ChildTable")
            self.doc.start_row()
            self.doc.start_cell("FGR-ParentHead", 4)
            self.doc.start_paragraph("FGR-ParentName")
            self.doc.write_text(self._("Children"))
            self.doc.end_paragraph()
            self.doc.end_cell()
            self.doc.end_row()
            index = 1
            for child_ref in family.get_child_ref_list():
                self.dump_child(index, child_ref.ref)
                index += 1
            self.doc.end_table()

        if self.recursive:
            for child_ref in family.get_child_ref_list():
                child = self.db.get_person_from_handle(child_ref.ref)
                for child_family_handle in child.get_family_handle_list():
                    if child_family_handle != family_handle:
                        self.doc.page_break()
                        self.dump_family(child_family_handle, (generation + 1))

    def write_report(self):
        flist = self.db.get_family_handles(sort_handles=True, locale=self._locale)
        if not self.filter:
            fam_list = flist
        else:
            fam_list = self.filter.apply(self.db, flist, user=self._user)
        if fam_list:
            with self._user.progress(
                _("Family Group Report"), _("Writing families"), len(fam_list)
            ) as step:
                for family_handle in fam_list:
                    self.dump_family(family_handle, 1)
                    self.doc.page_break()
                    step()
        else:
            raise ReportError(_("Empty report"), _("You did not specify anybody"))


# ------------------------------------------------------------------------
#
# FamilyGroupOptions
#
# ------------------------------------------------------------------------
class FamilyGroupOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__fid = None
        self.__filter = None
        self.__recursive = None
        self.__generations = None
        self._nf = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        return self.__filter.get_filter().get_name()

    def add_menu_options(self, menu):
        ##########################
        category_name = _("Report Options")
        add_option = partial(menu.add_option, category_name)
        ##########################

        self.__filter = FilterOption(_("Filter"), 0)
        self.__filter.set_help(_("Select the filter to be applied to the report."))
        add_option("filter", self.__filter)
        self.__filter.connect("value-changed", self.__filter_changed)

        self.__fid = FamilyOption(_("Center Family"))
        self.__fid.set_help(_("The center family for the filter"))
        add_option("family_id", self.__fid)
        self.__fid.connect("value-changed", self.__update_filters)

        self.__recursive = BooleanOption(_("Recursive (down)"), False)
        self.__recursive.set_help(
            _("Create reports for all descendants " "of this family.")
        )
        add_option("recursive", self.__recursive)
        self.__recursive.connect("value-changed", self.__recursive_changed)

        ##########################
        category_name = _("Report Options (2)")
        add_option = partial(menu.add_option, category_name)
        ##########################

        self._nf = stdoptions.add_name_format_option(menu, category_name)
        self._nf.connect("value-changed", self.__update_filters)

        self.__update_filters()

        stdoptions.add_place_format_option(menu, category_name)

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        locale_opt = stdoptions.add_localization_option(menu, category_name)

        stdoptions.add_date_format_option(menu, category_name, locale_opt)

        ##########################
        add_option = partial(menu.add_option, _("Include"))
        ##########################

        inc_par_mar = BooleanOption(_("Parent Marriage"), True)
        inc_par_mar.set_help(_("Whether to include marriage information for parents."))
        add_option("incParMar", inc_par_mar)

        inc_par_events = BooleanOption(_("Parent Events"), False)
        inc_par_events.set_help(_("Whether to include events for parents."))
        add_option("incParEvents", inc_par_events)

        inc_par_addr = BooleanOption(_("Parent Addresses"), False)
        inc_par_addr.set_help(_("Whether to include addresses for parents."))
        add_option("incParAddr", inc_par_addr)

        inc_par_notes = BooleanOption(_("Parent Notes"), False)
        inc_par_notes.set_help(_("Whether to include notes for parents."))
        add_option("incParNotes", inc_par_notes)

        incattrs = BooleanOption(_("Parent Attributes"), False)
        incattrs.set_help(_("Whether to include attributes."))
        add_option("incattrs", incattrs)

        inc_par_names = BooleanOption(_("Alternate Parent Names"), False)
        inc_par_names.set_help(_("Whether to include alternate names for parents."))
        add_option("incParNames", inc_par_names)

        ##########################
        category_name = _("Include (2)")
        add_option = partial(menu.add_option, category_name)
        ##########################

        stdoptions.add_gramps_id_option(menu, category_name)

        inc_fam_notes = BooleanOption(_("Family Notes"), False)
        inc_fam_notes.set_help(_("Whether to include notes for families."))
        add_option("incFamNotes", inc_fam_notes)

        inc_rel_dates = BooleanOption(_("Dates of Relatives"), False)
        inc_rel_dates.set_help(
            _("Whether to include dates for relatives " "(father, mother, spouse).")
        )
        add_option("incRelDates", inc_rel_dates)

        inc_chi_mar = BooleanOption(_("Children Marriages"), True)
        inc_chi_mar.set_help(_("Whether to include marriage information for children."))
        add_option("incChiMar", inc_chi_mar)

        self.__generations = BooleanOption(
            _("Generation numbers " "(recursive only)"), False
        )
        self.__generations.set_help(
            _("Whether to include the generation " "on each report (recursive only).")
        )
        add_option("generations", self.__generations)

        missinginfo = BooleanOption(_("Print fields for missing " "information"), True)
        missinginfo.set_help(_("Whether to include fields for missing " "information."))
        add_option("missinginfo", missinginfo)

    def __update_filters(self):
        """
        Update the filter list based on the selected family
        """
        fid = self.__fid.get_value()
        family = self.__db.get_family_from_gramps_id(fid)
        nfv = self._nf.get_value()
        filter_list = utils.get_family_filters(
            self.__db, family, include_single=True, name_format=nfv
        )
        self.__filter.set_filters(filter_list)

    def __filter_changed(self):
        """
        Handle filter change.
        If the filter is not family-specific, disable the family option
        """
        filter_value = self.__filter.get_value()
        if filter_value == 1:  # "Entire Database" (as "include_single=True")
            self.__fid.set_available(False)
        else:
            # The other filters need a center family (assume custom ones too)
            self.__fid.set_available(True)

        # only allow recursion if the center family is the only family
        if self.__recursive and filter_value == 0:
            self.__recursive.set_available(True)
        elif self.__recursive:
            self.__recursive.set_value(False)
            self.__recursive.set_available(False)

    def __recursive_changed(self):
        """
        Handle recursive change.
        """
        recursive_value = self.__recursive.get_value()
        self.__generations.set_available(recursive_value)

    def make_default_style(self, default_style):
        """Make default output style for the Family Group Report."""
        para = ParagraphStyle()
        # Paragraph Styles
        font = FontStyle()
        font.set_size(4)
        para.set_font(font)
        default_style.add_paragraph_style("FGR-blank", para)

        font = FontStyle()
        font.set_type_face(FONT_SANS_SERIF)
        font.set_size(16)
        font.set_bold(1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_header_level(1)
        para.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("FGR-Title", para)

        font = FontStyle()
        font.set_type_face(FONT_SERIF)
        font.set_size(10)
        font.set_bold(0)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_description(_("The basic style used for the text display."))
        default_style.add_paragraph_style("FGR-Normal", para)

        para = ParagraphStyle()
        font = FontStyle()
        font.set_type_face(FONT_SERIF)
        font.set_size(10)
        font.set_bold(0)
        para.set_font(font)
        para.set(lmargin=0.0)
        para.set_top_margin(0.0)
        para.set_bottom_margin(0.0)
        para.set_description(_("The basic style used for the note display."))
        default_style.add_paragraph_style("FGR-Note", para)

        font = FontStyle()
        font.set_type_face(FONT_SANS_SERIF)
        font.set_size(10)
        font.set_bold(1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_description(_("The style used for the text related to the children."))
        default_style.add_paragraph_style("FGR-ChildText", para)

        font = FontStyle()
        font.set_type_face(FONT_SANS_SERIF)
        font.set_size(12)
        font.set_bold(1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_header_level(3)
        para.set_description(_("The style used for the parent's name"))
        default_style.add_paragraph_style("FGR-ParentName", para)

        # Table Styles
        cell = TableCellStyle()
        cell.set_padding(0.2)
        cell.set_top_border(1)
        cell.set_bottom_border(1)
        cell.set_right_border(1)
        cell.set_left_border(1)
        default_style.add_cell_style("FGR-ParentHead", cell)

        cell = TableCellStyle()
        cell.set_padding(0.1)
        cell.set_bottom_border(1)
        cell.set_left_border(1)
        cell.set_right_border(1)  # for RTL
        default_style.add_cell_style("FGR-TextContents", cell)

        cell = TableCellStyle()
        cell.set_padding(0.1)
        cell.set_bottom_border(0)
        cell.set_left_border(1)
        cell.set_right_border(1)  # for RTL
        cell.set_padding(0.1)
        default_style.add_cell_style("FGR-TextChild1", cell)

        cell = TableCellStyle()
        cell.set_padding(0.1)
        cell.set_bottom_border(1)
        cell.set_left_border(1)
        cell.set_right_border(1)  # for RTL
        cell.set_padding(0.1)
        default_style.add_cell_style("FGR-TextChild2", cell)

        cell = TableCellStyle()
        cell.set_padding(0.1)
        cell.set_bottom_border(1)
        cell.set_right_border(1)
        cell.set_left_border(1)
        default_style.add_cell_style("FGR-TextContentsEnd", cell)

        cell = TableCellStyle()
        cell.set_padding(0.2)
        cell.set_bottom_border(1)
        cell.set_right_border(1)
        cell.set_left_border(1)
        default_style.add_cell_style("FGR-ChildName", cell)

        table = TableStyle()
        table.set_width(100)
        table.set_columns(3)
        table.set_column_width(0, 20)
        table.set_column_width(1, 40)
        table.set_column_width(2, 40)
        default_style.add_table_style("FGR-ParentTable", table)

        table = TableStyle()
        table.set_width(100)
        table.set_columns(4)
        table.set_column_width(0, 7)
        table.set_column_width(1, 18)
        table.set_column_width(2, 35)
        table.set_column_width(3, 40)
        default_style.add_table_style("FGR-ChildTable", table)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2004  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import csv

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.config import config
from .tabbeddoc import *
from ...constfunc import win

_ = glocale.translation.gettext


class CSVTab(TabbedDoc):
    def __init__(self, columns):
        TabbedDoc.__init__(self, columns)
        self.filename = None
        self.f = None
        self.dlist = []
        self.writer = None

    def open(self, filename):
        if filename[-4:] != ".csv":
            self.filename = filename + ".csv"
        else:
            self.filename = filename

        self.f = open(
            self.filename, "w", newline="", encoding="utf_8_sig" if win() else "utf_8"
        )
        my_dialect = config.get("csv.dialect")
        my_delimiter = config.get("csv.delimiter")
        if my_dialect == _("Custom"):
            self.writer = csv.writer(self.f, delimiter=my_delimiter)
        else:
            self.writer = csv.writer(self.f, dialect=my_dialect)

    def close(self):
        assert self.f
        self.f.close()

    def start_row(self):
        self.dlist = []

    def end_row(self):
        self.writer.writerow(self.dlist)

    def write_cell(self, text):
        self.dlist.append(text)

    def start_page(self):
        pass

    def end_page(self):
        pass


if __name__ == "__main__":
    file = CSVTab(2, 3)
    file.open("test.csv")
    file.start_page()
    for i in [("one", "two", "three"), ('fo"ur', "fi,ve", "six")]:
        file.start_row()
        for j in i:
            file.write_cell(j)
        file.end_row()
    file.end_page()
    file.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Utilities/Generate SoundEx Codes"""

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.soundex import soundex
from gramps.gui.autocomp import fill_combo
from gramps.gen.plug import Gramplet
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# SoundGen
#
# -------------------------------------------------------------------------
class SoundGen(Gramplet):
    """
    Generates SoundEx codes.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)

    def build_gui(self):
        """
        Build the GUI interface.
        """
        grid = Gtk.Grid()
        grid.set_border_width(6)
        grid.set_row_spacing(6)
        grid.set_column_spacing(20)
        label1 = Gtk.Label(label=_("Name:"), halign=Gtk.Align.START)
        grid.attach(label1, 0, 0, 1, 1)
        label2 = Gtk.Label(label=_("SoundEx code:"), halign=Gtk.Align.START)
        grid.attach(label2, 0, 1, 1, 1)
        self.autocomp = Gtk.ComboBox.new_with_entry()
        grid.attach(self.autocomp, 1, 0, 1, 1)
        self.value = Gtk.Label(halign=Gtk.Align.START)
        self.value.set_selectable(True)
        grid.attach(self.value, 1, 1, 1, 1)
        self.name = self.autocomp.get_child()

        self.name.connect("changed", self.on_apply_clicked)

        grid.show_all()
        return grid

    def db_changed(self):
        if not self.dbstate.is_open():
            return

        names = []
        person = None
        for person in self.dbstate.db.iter_people():
            lastname = person.get_primary_name().get_surname()
            if lastname not in names:
                names.append(lastname)

        names.sort()

        fill_combo(self.autocomp, names)

        if person:
            n = person.get_primary_name().get_surname()
            self.name.set_text(n)
            try:
                se_text = soundex(n)
            except UnicodeEncodeError:
                se_text = soundex("")
            self.value.set_text(se_text)
        else:
            self.name.set_text("")

    def on_apply_clicked(self, obj):
        try:
            se_text = soundex(str(obj.get_text()))
        except UnicodeEncodeError:
            se_text = soundex("")
        self.value.set_text(se_text)
